"""
Rösch-MacAdam colour solid - Visible Spectrum Volume Computations
=================================================================

Defines the objects related to *Rösch-MacAdam* colour solid, visible spectrum
volume computations.

References
----------
-   :cite:`Lindbloom2015` : Lindbloom, B. (2015). About the Lab Gamut.
    Retrieved August 20, 2018, from
    http://www.brucelindbloom.com/LabGamutDisplayHelp.html
-   :cite:`Mansencal2018` : Mansencal, T. (2018). How is the visible gamut
    bounded? Retrieved August 19, 2018, from
    https://stackoverflow.com/a/48396021/931625
-   :cite:`Martinez-Verdu2007` : Martínez-Verdú, F., Perales, E., Chorro, E.,
    de Fez, D., Viqueira, V., & Gilabert, E. (2007). Computation and
    visualization of the MacAdam limits for any lightness, hue angle, and light
    source. Journal of the Optical Society of America A, 24(6), 1501.
    doi:10.1364/JOSAA.24.001501
"""

from __future__ import annotations

import numpy as np

from colour.colorimetry import (
    MultiSpectralDistributions,
    SpectralDistribution,
    SpectralShape,
    handle_spectral_arguments,
    msds_to_XYZ,
)
from colour.constants import DEFAULT_FLOAT_DTYPE
from colour.hints import (
    Any,
    ArrayLike,
    Boolean,
    Dict,
    Floating,
    Integer,
    Literal,
    NDArray,
    Optional,
    Union,
)
from colour.volume import is_within_mesh_volume
from colour.utilities import CACHE_REGISTRY, zeros, validate_method

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "SPECTRAL_SHAPE_OUTER_SURFACE_XYZ",
    "generate_pulse_waves",
    "XYZ_outer_surface",
    "solid_RoschMacAdam",
    "is_within_visible_spectrum",
]

SPECTRAL_SHAPE_OUTER_SURFACE_XYZ: SpectralShape = SpectralShape(360, 780, 5)
"""
Default spectral shape according to *ASTM E308-15* practise shape but using an
interval of 5.
"""

_CACHE_OUTER_SURFACE_XYZ: Dict = CACHE_REGISTRY.register_cache(
    f"{__name__}._CACHE_OUTER_SURFACE_XYZ"
)

_CACHE_OUTER_SURFACE_XYZ_POINTS: Dict = CACHE_REGISTRY.register_cache(
    f"{__name__}._CACHE_OUTER_SURFACE_XYZ_POINTS"
)


def generate_pulse_waves(
    bins: Integer,
    pulse_order: Union[Literal["Bins", "Pulse Wave Width"], str] = "Bins",
    filter_jagged_pulses: Boolean = False,
) -> NDArray:
    """
    Generate the pulse waves of given number of bins necessary to totally
    stimulate the colour matching functions and produce the *Rösch-MacAdam*
    colour solid.

    Assuming 5 bins, a first set of SPDs would be as follows::

        1 0 0 0 0
        0 1 0 0 0
        0 0 1 0 0
        0 0 0 1 0
        0 0 0 0 1

    The second one::

        1 1 0 0 0
        0 1 1 0 0
        0 0 1 1 0
        0 0 0 1 1
        1 0 0 0 1

    The third:

        1 1 1 0 0
        0 1 1 1 0
        0 0 1 1 1
        1 0 0 1 1
        1 1 0 0 1

    Etc...

    Parameters
    ----------
    bins
        Number of bins of the pulse waves.
    pulse_order
        Method for ordering the pulse waves. *Bins* is the default order, with
        *Pulse Wave Width* ordering, instead of iterating over the pulse wave
        widths first, iteration occurs over the bins, producing blocks of pulse
        waves with increasing width.
    filter_jagged_pulses
        Whether to filter jagged pulses. When ``pulse_order`` is set to
        *Pulse Wave Width*, the pulses are ordered by increasing width. Because
        of the discrete nature of the underlying signal, the resulting pulses
        will be jagged. For example assuming 5 bins, the center block with
        the two extreme values added would be as follows::

            0 0 0 0 0
            0 0 1 0 0
            0 0 1 1 0 <--
            0 1 1 1 0
            0 1 1 1 1 <--
            1 1 1 1 1

        Setting the ``filter_jagged_pulses`` parameter to `True` will result
        in the removal of the two marked pulse waves above thus avoiding jagged
        lines when plotting and having to resort to excessive ``bins`` values.

    Returns
    -------
    :class:`numpy.ndarray`
        Pulse waves.

    References
    ----------
    :cite:`Lindbloom2015`, :cite:`Mansencal2018`, :cite:`Martinez-Verdu2007`

    Examples
    --------
    >>> generate_pulse_waves(5)
    array([[ 0.,  0.,  0.,  0.,  0.],
           [ 1.,  0.,  0.,  0.,  0.],
           [ 0.,  1.,  0.,  0.,  0.],
           [ 0.,  0.,  1.,  0.,  0.],
           [ 0.,  0.,  0.,  1.,  0.],
           [ 0.,  0.,  0.,  0.,  1.],
           [ 1.,  1.,  0.,  0.,  0.],
           [ 0.,  1.,  1.,  0.,  0.],
           [ 0.,  0.,  1.,  1.,  0.],
           [ 0.,  0.,  0.,  1.,  1.],
           [ 1.,  0.,  0.,  0.,  1.],
           [ 1.,  1.,  1.,  0.,  0.],
           [ 0.,  1.,  1.,  1.,  0.],
           [ 0.,  0.,  1.,  1.,  1.],
           [ 1.,  0.,  0.,  1.,  1.],
           [ 1.,  1.,  0.,  0.,  1.],
           [ 1.,  1.,  1.,  1.,  0.],
           [ 0.,  1.,  1.,  1.,  1.],
           [ 1.,  0.,  1.,  1.,  1.],
           [ 1.,  1.,  0.,  1.,  1.],
           [ 1.,  1.,  1.,  0.,  1.],
           [ 1.,  1.,  1.,  1.,  1.]])
    >>> generate_pulse_waves(5, 'Pulse Wave Width')
    array([[ 0.,  0.,  0.,  0.,  0.],
           [ 1.,  0.,  0.,  0.,  0.],
           [ 1.,  1.,  0.,  0.,  0.],
           [ 1.,  1.,  0.,  0.,  1.],
           [ 1.,  1.,  1.,  0.,  1.],
           [ 0.,  1.,  0.,  0.,  0.],
           [ 0.,  1.,  1.,  0.,  0.],
           [ 1.,  1.,  1.,  0.,  0.],
           [ 1.,  1.,  1.,  1.,  0.],
           [ 0.,  0.,  1.,  0.,  0.],
           [ 0.,  0.,  1.,  1.,  0.],
           [ 0.,  1.,  1.,  1.,  0.],
           [ 0.,  1.,  1.,  1.,  1.],
           [ 0.,  0.,  0.,  1.,  0.],
           [ 0.,  0.,  0.,  1.,  1.],
           [ 0.,  0.,  1.,  1.,  1.],
           [ 1.,  0.,  1.,  1.,  1.],
           [ 0.,  0.,  0.,  0.,  1.],
           [ 1.,  0.,  0.,  0.,  1.],
           [ 1.,  0.,  0.,  1.,  1.],
           [ 1.,  1.,  0.,  1.,  1.],
           [ 1.,  1.,  1.,  1.,  1.]])
    >>> generate_pulse_waves(5, 'Pulse Wave Width', True)
    array([[ 0.,  0.,  0.,  0.,  0.],
           [ 1.,  0.,  0.,  0.,  0.],
           [ 1.,  1.,  0.,  0.,  1.],
           [ 0.,  1.,  0.,  0.,  0.],
           [ 1.,  1.,  1.,  0.,  0.],
           [ 0.,  0.,  1.,  0.,  0.],
           [ 0.,  1.,  1.,  1.,  0.],
           [ 0.,  0.,  0.,  1.,  0.],
           [ 0.,  0.,  1.,  1.,  1.],
           [ 0.,  0.,  0.,  0.,  1.],
           [ 1.,  0.,  0.,  1.,  1.],
           [ 1.,  1.,  1.,  1.,  1.]])
    """

    pulse_order = validate_method(
        pulse_order,
        ["Bins", "Pulse Wave Width"],
        '"{0}" pulse order is invalid, it must be one of {1}!',
    )

    square_waves = []
    square_waves_basis = np.tril(
        np.ones((bins, bins), dtype=DEFAULT_FLOAT_DTYPE)
    )[0:-1, :]

    if pulse_order.lower() == "bins":
        for square_wave_basis in square_waves_basis:
            for i in range(bins):
                square_waves.append(np.roll(square_wave_basis, i))
    else:
        for i in range(bins):
            for j, square_wave_basis in enumerate(square_waves_basis):
                square_waves.append(np.roll(square_wave_basis, i - j // 2))

        if filter_jagged_pulses:
            square_waves = square_waves[::2]

    return np.vstack(
        [
            zeros(bins),
            np.vstack(square_waves),
            np.ones(bins, dtype=DEFAULT_FLOAT_DTYPE),
        ]
    )


def XYZ_outer_surface(
    cmfs: Optional[MultiSpectralDistributions] = None,
    illuminant: Optional[SpectralDistribution] = None,
    point_order: Union[Literal["Bins", "Pulse Wave Width"], str] = "Bins",
    filter_jagged_points: Boolean = False,
    **kwargs: Any,
) -> NDArray:
    """
    Generate the *Rösch-MacAdam* colour solid, i.e. *CIE XYZ* colourspace
    outer surface, for given colour matching functions using multi-spectral
    conversion of pulse waves to *CIE XYZ* tristimulus values.

    Parameters
    ----------
    cmfs
        Standard observer colour matching functions, default to the
        *CIE 1931 2 Degree Standard Observer*.
    illuminant
        Illuminant spectral distribution, default to *CIE Illuminant E*.
    point_order
        Method for ordering the underlying pulse waves used to generate the
        *Rösch-MacAdam* colour solid. *Bins* is the default order, with
        *Pulse Wave Width* ordering, instead of iterating over the pulse wave
        widths first, iteration occurs over the bins, producing blocks of pulse
        waves with increasing width.
    filter_jagged_points
        Whether to filter the underlying jagged pulses. When ``point_order`` is
        set to *Pulse Wave Width*, the pulses are ordered by increasing width.
        Because of the discrete nature of the underlying signal, the resulting
        pulses will be jagged. For example assuming 5 bins, the center block
        with the two extreme values added would be as follows::

            0 0 0 0 0
            0 0 1 0 0
            0 0 1 1 0 <--
            0 1 1 1 0
            0 1 1 1 1 <--
            1 1 1 1 1

        Setting the ``filter_jagged_points`` parameter to `True` will result
        in the removal of the two marked pulse waves above thus avoiding jagged
        lines when plotting and having to resort to excessive ``bins`` values.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.msds_to_XYZ`},
        See the documentation of the previously listed definition.

    Returns
    -------
    :class:`numpy.ndarray`
        *Rösch-MacAdam* colour solid, *CIE XYZ* outer surface tristimulus
        values.

    References
    ----------
    :cite:`Lindbloom2015`, :cite:`Mansencal2018`, :cite:`Martinez-Verdu2007`

    Examples
    --------
    >>> from colour import MSDS_CMFS, SPECTRAL_SHAPE_DEFAULT
    >>> shape = SpectralShape(
    ...     SPECTRAL_SHAPE_DEFAULT.start, SPECTRAL_SHAPE_DEFAULT.end, 84
    ... )
    >>> cmfs = MSDS_CMFS['CIE 1931 2 Degree Standard Observer']
    >>> XYZ_outer_surface(cmfs.copy().align(shape))  # doctest: +ELLIPSIS
    array([[  0.0000000...e+00,   0.0000000...e+00,   0.0000000...e+00],
           [  9.6361381...e-05,   2.9056776...e-06,   4.4961226...e-04],
           [  2.5910529...e-01,   2.1031298...e-02,   1.3207468...e+00],
           [  1.0561021...e-01,   6.2038243...e-01,   3.5423571...e-02],
           [  7.2647980...e-01,   3.5460869...e-01,   2.1005149...e-04],
           [  1.0971874...e-02,   3.9635453...e-03,   0.0000000...e+00],
           [  3.0792572...e-05,   1.1119762...e-05,   0.0000000...e+00],
           [  2.5920165...e-01,   2.1034203...e-02,   1.3211965...e+00],
           [  3.6471551...e-01,   6.4141373...e-01,   1.3561704...e+00],
           [  8.3209002...e-01,   9.7499113...e-01,   3.5633622...e-02],
           [  7.3745167...e-01,   3.5857224...e-01,   2.1005149...e-04],
           [  1.1002667...e-02,   3.9746651...e-03,   0.0000000...e+00],
           [  1.2715395...e-04,   1.4025439...e-05,   4.4961226...e-04],
           [  3.6481187...e-01,   6.4141663...e-01,   1.3566200...e+00],
           [  1.0911953...e+00,   9.9602242...e-01,   1.3563805...e+00],
           [  8.4306189...e-01,   9.7895467...e-01,   3.5633622...e-02],
           [  7.3748247...e-01,   3.5858336...e-01,   2.1005149...e-04],
           [  1.1099028...e-02,   3.9775708...e-03,   4.4961226...e-04],
           [  2.5923244...e-01,   2.1045323...e-02,   1.3211965...e+00],
           [  1.0912916...e+00,   9.9602533...e-01,   1.3568301...e+00],
           [  1.1021671...e+00,   9.9998597...e-01,   1.3563805...e+00],
           [  8.4309268...e-01,   9.7896579...e-01,   3.5633622...e-02],
           [  7.3757883...e-01,   3.5858626...e-01,   6.5966375...e-04],
           [  2.7020432...e-01,   2.5008868...e-02,   1.3211965...e+00],
           [  3.6484266...e-01,   6.4142775...e-01,   1.3566200...e+00],
           [  1.1022635...e+00,   9.9998888...e-01,   1.3568301...e+00],
           [  1.1021979...e+00,   9.9999709...e-01,   1.3563805...e+00],
           [  8.4318905...e-01,   9.7896870...e-01,   3.6083235...e-02],
           [  9.9668412...e-01,   3.7961756...e-01,   1.3214065...e+00],
           [  3.7581454...e-01,   6.4539130...e-01,   1.3566200...e+00],
           [  1.0913224...e+00,   9.9603645...e-01,   1.3568301...e+00],
           [  1.1022943...e+00,   1.0000000...e+00,   1.3568301...e+00]])
    """

    cmfs, illuminant = handle_spectral_arguments(
        cmfs,
        illuminant,
        "CIE 1931 2 Degree Standard Observer",
        "E",
        SPECTRAL_SHAPE_OUTER_SURFACE_XYZ,
    )

    settings = dict(kwargs)
    settings.update({"shape": cmfs.shape})

    key = (
        hash(cmfs),
        hash(illuminant),
        point_order,
        filter_jagged_points,
        str(settings),
    )
    XYZ = _CACHE_OUTER_SURFACE_XYZ.get(key)

    if XYZ is None:
        pulse_waves = generate_pulse_waves(
            len(cmfs.wavelengths), point_order, filter_jagged_points
        )
        XYZ = (
            msds_to_XYZ(
                pulse_waves, cmfs, illuminant, method="Integration", **settings
            )
            / 100
        )

        _CACHE_OUTER_SURFACE_XYZ[key] = XYZ

    return XYZ


solid_RoschMacAdam = XYZ_outer_surface


def is_within_visible_spectrum(
    XYZ: ArrayLike,
    cmfs: Optional[MultiSpectralDistributions] = None,
    illuminant: Optional[SpectralDistribution] = None,
    tolerance: Optional[Floating] = None,
    **kwargs: Any,
) -> NDArray:
    """
    Return whether given *CIE XYZ* tristimulus values are within the visible
    spectrum volume, i.e. *Rösch-MacAdam* colour solid, for given colour
    matching functions and illuminant.

    Parameters
    ----------
    XYZ
        *CIE XYZ* tristimulus values.
    cmfs
        Standard observer colour matching functions, default to the
        *CIE 1931 2 Degree Standard Observer*.
    illuminant
        Illuminant spectral distribution, default to *CIE Illuminant E*.
    tolerance
        Tolerance allowed in the inside-triangle check.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.msds_to_XYZ`},
        See the documentation of the previously listed definition.

    Returns
    -------
    :class:`numpy.ndarray`
        Are *CIE XYZ* tristimulus values within the visible spectrum volume,
        i.e. *Rösch-MacAdam* colour solid.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    Examples
    --------
    >>> import numpy as np
    >>> is_within_visible_spectrum(np.array([0.3205, 0.4131, 0.51]))
    array(True, dtype=bool)
    >>> a = np.array([[0.3205, 0.4131, 0.51],
    ...               [-0.0005, 0.0031, 0.001]])
    >>> is_within_visible_spectrum(a)
    array([ True, False], dtype=bool)
    """

    cmfs, illuminant = handle_spectral_arguments(
        cmfs,
        illuminant,
        "CIE 1931 2 Degree Standard Observer",
        "E",
        SPECTRAL_SHAPE_OUTER_SURFACE_XYZ,
    )

    key = (hash(cmfs), hash(illuminant), str(kwargs))
    vertices = _CACHE_OUTER_SURFACE_XYZ_POINTS.get(key)

    if vertices is None:
        _CACHE_OUTER_SURFACE_XYZ_POINTS[key] = vertices = solid_RoschMacAdam(
            cmfs, illuminant, **kwargs
        )

    return is_within_mesh_volume(XYZ, vertices, tolerance)



"""
Sony Colourspaces
=================

Defines the *Sony* colourspaces:

-   :attr:`colour.models.RGB_COLOURSPACE_S_GAMUT`.
-   :attr:`colour.models.RGB_COLOURSPACE_S_GAMUT3`.
-   :attr:`colour.models.RGB_COLOURSPACE_S_GAMUT3_CINE`.
-   :attr:`colour.models.RGB_COLOURSPACE_VENICE_S_GAMUT3`.
-   :attr:`colour.models.RGB_COLOURSPACE_VENICE_S_GAMUT3_CINE`.

Notes
-----
-   The *Venice S-Gamut3* and *Venice S-Gamut3.Cine* primaries and whitepoint
    were derived with the following `Google Colab Notebook \
<https://colab.research.google.com/drive/1ZGTij7jT8eZRMPUkyWlv_x5ix5Q5twMB>`__.

References
----------
-   :cite:`Gaggioni` : Gaggioni, H., Dhanendra, P., Yamashita, J., Kawada, N.,
    Endo, K., & Clark, C. (n.d.). S-Log: A new LUT for digital production
    mastering and interchange applications (Vol. 709, pp. 1-13).
    http://pro.sony.com/bbsccms/assets/files/mkt/cinema/solutions/slog_manual.pdf
-   :cite:`SonyCorporation` : Sony Corporation. (n.d.). S-Log Whitepaper (pp.
    1-17). http://www.theodoropoulos.info/attachments/076_on%20S-Log.pdf
-   :cite:`SonyCorporationd` : Sony Corporation. (n.d.). Technical Summary
    for S-Gamut3.Cine/S-Log3 and S-Gamut3/S-Log3 (pp. 1-7).
    http://community.sony.com/sony/attachments/sony/\
large-sensor-camera-F5-F55/12359/2/\
TechnicalSummary_for_S-Gamut3Cine_S-Gamut3_S-Log3_V1_00.pdf
-   :cite:`SonyCorporatione` : Sony Corporation. (n.d.).
    S-Gamut3_S-Gamut3Cine_Matrix.xlsx.
    https://community.sony.com/sony/attachments/sony/\
large-sensor-camera-F5-F55/12359/3/S-Gamut3_S-Gamut3Cine_Matrix.xlsx
-   :cite:`SonyElectronicsCorporation2020` : Sony Electronics Corporation.
    (2020). IDT.Sony.Venice_SLog3_SGamut3.ctl. https://github.com/ampas/\
aces-dev/blob/710ecbe52c87ce9f4a1e02c8ddf7ea0d6b611cc8/transforms/ctl/idt/\
vendorSupplied/sony/IDT.Sony.Venice_SLog3_SGamut3.ctl
-   :cite:`SonyElectronicsCorporation2020a` : Sony Electronics Corporation.
    (2020). IDT.Sony.Venice_SLog3_SGamut3Cine.ctl. https://github.com/ampas/\
aces-dev/blob/710ecbe52c87ce9f4a1e02c8ddf7ea0d6b611cc8/transforms/ctl/idt/\
vendorSupplied/sony/IDT.Sony.Venice_SLog3_SGamut3Cine.ctl
"""

from __future__ import annotations

import numpy as np

from colour.colorimetry import CCS_ILLUMINANTS
from colour.hints import NDArray
from colour.models.rgb import (
    RGB_Colourspace,
    log_encoding_SLog2,
    log_decoding_SLog2,
    log_encoding_SLog3,
    log_decoding_SLog3,
    normalised_primary_matrix,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "PRIMARIES_S_GAMUT",
    "WHITEPOINT_NAME_S_GAMUT",
    "CCS_WHITEPOINT_S_GAMUT",
    "MATRIX_S_GAMUT_TO_XYZ",
    "MATRIX_XYZ_TO_S_GAMUT",
    "RGB_COLOURSPACE_S_GAMUT",
    "PRIMARIES_S_GAMUT3",
    "WHITEPOINT_NAME_S_GAMUT3",
    "CCS_WHITEPOINT_S_GAMUT3",
    "MATRIX_S_GAMUT3_TO_XYZ",
    "MATRIX_XYZ_TO_S_GAMUT3",
    "RGB_COLOURSPACE_S_GAMUT3",
    "PRIMARIES_S_GAMUT3_CINE",
    "WHITEPOINT_NAME_S_GAMUT3_CINE",
    "CCS_WHITEPOINT_S_GAMUT3_CINE",
    "MATRIX_S_GAMUT3_CINE_TO_XYZ",
    "MATRIX_XYZ_TO_S_GAMUT3_CINE",
    "RGB_COLOURSPACE_S_GAMUT3_CINE",
    "PRIMARIES_VENICE_S_GAMUT3",
    "WHITEPOINT_NAME_VENICE_S_GAMUT3",
    "CCS_WHITEPOINT_VENICE_S_GAMUT3",
    "MATRIX_VENICE_S_GAMUT3_TO_XYZ",
    "MATRIX_XYZ_TO_VENICE_S_GAMUT3",
    "RGB_COLOURSPACE_VENICE_S_GAMUT3",
    "PRIMARIES_VENICE_S_GAMUT3_CINE",
    "WHITEPOINT_NAME_VENICE_S_GAMUT3_CINE",
    "CCS_WHITEPOINT_VENICE_S_GAMUT3_CINE",
    "MATRIX_VENICE_S_GAMUT3_CINE_TO_XYZ",
    "MATRIX_XYZ_TO_VENICE_S_GAMUT3_CINE",
    "RGB_COLOURSPACE_VENICE_S_GAMUT3_CINE",
]

PRIMARIES_S_GAMUT: NDArray = np.array(
    [
        [0.7300, 0.2800],
        [0.1400, 0.8550],
        [0.1000, -0.0500],
    ]
)
"""*S-Gamut* colourspace primaries."""

WHITEPOINT_NAME_S_GAMUT: str = "D65"
"""*S-Gamut* colourspace whitepoint name."""

CCS_WHITEPOINT_S_GAMUT: NDArray = CCS_ILLUMINANTS[
    "CIE 1931 2 Degree Standard Observer"
][WHITEPOINT_NAME_S_GAMUT]
"""*S-Gamut* colourspace whitepoint chromaticity coordinates."""

MATRIX_S_GAMUT_TO_XYZ: NDArray = np.array(
    [
        [0.7064827132, 0.1288010498, 0.1151721641],
        [0.2709796708, 0.7866064112, -0.0575860820],
        [-0.0096778454, 0.0046000375, 1.0941355587],
    ]
)
"""*S-Gamut* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_S_GAMUT: NDArray = np.array(
    [
        [1.5073998991, -0.2458221374, -0.1716116808],
        [-0.5181517271, 1.3553912409, 0.1258786682],
        [0.0155116982, -0.0078727714, 0.9119163656],
    ]
)
"""*CIE XYZ* tristimulus values to *S-Gamut* colourspace matrix."""

RGB_COLOURSPACE_S_GAMUT: RGB_Colourspace = RGB_Colourspace(
    "S-Gamut",
    PRIMARIES_S_GAMUT,
    CCS_WHITEPOINT_S_GAMUT,
    WHITEPOINT_NAME_S_GAMUT,
    MATRIX_S_GAMUT_TO_XYZ,
    MATRIX_XYZ_TO_S_GAMUT,
    log_encoding_SLog2,
    log_decoding_SLog2,
)
RGB_COLOURSPACE_S_GAMUT.__doc__ = """
*S-Gamut* colourspace.

References
----------
:cite:`Gaggioni`, :cite:`SonyCorporation`
"""

PRIMARIES_S_GAMUT3: NDArray = PRIMARIES_S_GAMUT
"""*S-Gamut3* colourspace primaries."""

WHITEPOINT_NAME_S_GAMUT3: str = WHITEPOINT_NAME_S_GAMUT
"""*S-Gamut3* colourspace whitepoint name."""

CCS_WHITEPOINT_S_GAMUT3: NDArray = CCS_WHITEPOINT_S_GAMUT
"""*S-Gamut3* colourspace whitepoint chromaticity coordinates."""

MATRIX_S_GAMUT3_TO_XYZ: NDArray = MATRIX_S_GAMUT_TO_XYZ
"""*S-Gamut3* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_S_GAMUT3: NDArray = MATRIX_XYZ_TO_S_GAMUT
"""*CIE XYZ* tristimulus values to *S-Gamut3* colourspace matrix."""

RGB_COLOURSPACE_S_GAMUT3: RGB_Colourspace = RGB_Colourspace(
    "S-Gamut3",
    PRIMARIES_S_GAMUT3,
    CCS_WHITEPOINT_S_GAMUT3,
    WHITEPOINT_NAME_S_GAMUT3,
    MATRIX_S_GAMUT3_TO_XYZ,
    MATRIX_XYZ_TO_S_GAMUT3,
    log_encoding_SLog3,
    log_decoding_SLog3,
)
RGB_COLOURSPACE_S_GAMUT3.__doc__ = """
*S-Gamut3* colourspace.

References
----------
:cite:`SonyCorporationd`
"""

PRIMARIES_S_GAMUT3_CINE: NDArray = np.array(
    [
        [0.76600, 0.27500],
        [0.22500, 0.80000],
        [0.08900, -0.08700],
    ]
)
"""*S-Gamut3.Cine* colourspace primaries."""

WHITEPOINT_NAME_S_GAMUT3_CINE: str = WHITEPOINT_NAME_S_GAMUT
"""*S-Gamut3.Cine* colourspace whitepoint name."""

CCS_WHITEPOINT_S_GAMUT3_CINE: NDArray = CCS_WHITEPOINT_S_GAMUT
"""*S-Gamut3.Cine* colourspace whitepoint chromaticity coordinates."""

MATRIX_S_GAMUT3_CINE_TO_XYZ: NDArray = np.array(
    [
        [0.5990839208, 0.2489255161, 0.1024464902],
        [0.2150758201, 0.8850685017, -0.1001443219],
        [-0.0320658495, -0.0276583907, 1.1487819910],
    ]
)
"""*S-Gamut3.Cine* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_S_GAMUT3_CINE: NDArray = np.array(
    [
        [1.8467789693, -0.5259861230, -0.2105452114],
        [-0.4441532629, 1.2594429028, 0.1493999729],
        [0.0408554212, 0.0156408893, 0.8682072487],
    ]
)
"""*CIE XYZ* tristimulus values to *S-Gamut3.Cine* colourspace matrix."""

RGB_COLOURSPACE_S_GAMUT3_CINE: RGB_Colourspace = RGB_Colourspace(
    "S-Gamut3.Cine",
    PRIMARIES_S_GAMUT3_CINE,
    CCS_WHITEPOINT_S_GAMUT3_CINE,
    WHITEPOINT_NAME_S_GAMUT3_CINE,
    MATRIX_S_GAMUT3_CINE_TO_XYZ,
    MATRIX_XYZ_TO_S_GAMUT3_CINE,
    log_encoding_SLog3,
    log_decoding_SLog3,
)
RGB_COLOURSPACE_S_GAMUT3_CINE.__doc__ = """
*S-Gamut3.Cine* colourspace.

References
----------
:cite:`SonyCorporatione`
"""

PRIMARIES_VENICE_S_GAMUT3: NDArray = np.array(
    [
        [0.740464264304292, 0.279364374750660],
        [0.089241145423286, 0.893809528608105],
        [0.110488236673827, -0.052579333080476],
    ]
)
"""*Venice S-Gamut3* colourspace primaries."""

WHITEPOINT_NAME_VENICE_S_GAMUT3: str = WHITEPOINT_NAME_S_GAMUT
"""*Venice S-Gamut3* colourspace whitepoint name."""

CCS_WHITEPOINT_VENICE_S_GAMUT3: NDArray = CCS_WHITEPOINT_S_GAMUT
"""*Venice S-Gamut3* colourspace whitepoint chromaticity coordinates."""

MATRIX_VENICE_S_GAMUT3_TO_XYZ: NDArray = normalised_primary_matrix(
    PRIMARIES_VENICE_S_GAMUT3, CCS_WHITEPOINT_VENICE_S_GAMUT3
)
"""*Venice S-Gamut3* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_VENICE_S_GAMUT3: NDArray = np.linalg.inv(
    MATRIX_VENICE_S_GAMUT3_TO_XYZ
)
"""*CIE XYZ* tristimulus values to *Venice S-Gamut3* colourspace matrix."""

RGB_COLOURSPACE_VENICE_S_GAMUT3: RGB_Colourspace = RGB_Colourspace(
    "Venice S-Gamut3",
    PRIMARIES_VENICE_S_GAMUT3,
    CCS_WHITEPOINT_VENICE_S_GAMUT3,
    WHITEPOINT_NAME_VENICE_S_GAMUT3,
    MATRIX_VENICE_S_GAMUT3_TO_XYZ,
    MATRIX_XYZ_TO_VENICE_S_GAMUT3,
    log_encoding_SLog3,
    log_decoding_SLog3,
)
RGB_COLOURSPACE_VENICE_S_GAMUT3.__doc__ = """
*Venice S-Gamut3* colourspace.

References
----------
:cite:`SonyElectronicsCorporation2020`
"""

PRIMARIES_VENICE_S_GAMUT3_CINE: NDArray = np.array(
    [
        [0.775901871567345, 0.274502392854799],
        [0.188682902773355, 0.828684937020288],
        [0.101337382499301, -0.089187517306263],
    ]
)
"""*Venice S-Gamut3.Cine* colourspace primaries."""

WHITEPOINT_NAME_VENICE_S_GAMUT3_CINE: str = WHITEPOINT_NAME_S_GAMUT
"""*Venice S-Gamut3.Cine* colourspace whitepoint name."""

CCS_WHITEPOINT_VENICE_S_GAMUT3_CINE: NDArray = CCS_WHITEPOINT_S_GAMUT
"""*Venice S-Gamut3.Cine* colourspace whitepoint chromaticity coordinates."""

MATRIX_VENICE_S_GAMUT3_CINE_TO_XYZ: NDArray = normalised_primary_matrix(
    PRIMARIES_VENICE_S_GAMUT3_CINE, CCS_WHITEPOINT_VENICE_S_GAMUT3_CINE
)
"""*Venice S-Gamut3.Cine* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_VENICE_S_GAMUT3_CINE: NDArray = np.linalg.inv(
    MATRIX_VENICE_S_GAMUT3_CINE_TO_XYZ
)
"""*CIE XYZ* tristimulus values to *Venice S-Gamut3.Cine* colourspace matrix."""

RGB_COLOURSPACE_VENICE_S_GAMUT3_CINE: RGB_Colourspace = RGB_Colourspace(
    "Venice S-Gamut3.Cine",
    PRIMARIES_VENICE_S_GAMUT3_CINE,
    CCS_WHITEPOINT_VENICE_S_GAMUT3_CINE,
    WHITEPOINT_NAME_VENICE_S_GAMUT3_CINE,
    MATRIX_VENICE_S_GAMUT3_CINE_TO_XYZ,
    MATRIX_XYZ_TO_VENICE_S_GAMUT3_CINE,
    log_encoding_SLog3,
    log_decoding_SLog3,
)
RGB_COLOURSPACE_VENICE_S_GAMUT3_CINE.__doc__ = """
*Venice S-Gamut3.Cine* colourspace.

References
----------
:cite:`SonyElectronicsCorporation2020a`
"""



from .aces_it import MSDS_ACES_RICD
from .cameras import MSDS_CAMERA_SENSITIVITIES
from .colour_checkers import (
    CCS_COLOURCHECKERS,
    ColourChecker,
    SDS_COLOURCHECKERS,
)
from .displays import MSDS_DISPLAY_PRIMARIES
from .filters import SDS_FILTERS
from .lenses import SDS_LENSES

__all__ = [
    "MSDS_ACES_RICD",
]
__all__ += [
    "MSDS_CAMERA_SENSITIVITIES",
]
__all__ += [
    "CCS_COLOURCHECKERS",
    "ColourChecker",
    "SDS_COLOURCHECKERS",
]
__all__ += [
    "MSDS_DISPLAY_PRIMARIES",
]
__all__ += [
    "SDS_FILTERS",
]
__all__ += [
    "SDS_LENSES",
]



"""
CIE RGB Colourspace
===================

Defines the *CIE RGB* colourspace:

-   :attr:`colour.models.RGB_COLOURSPACE_CIE_RGB`.

References
----------
-   :cite:`Fairman1997` : Fairman, H. S., Brill, M. H., & Hemmendinger, H.
    (1997). How the CIE 1931 color-matching functions were derived from
    Wright-Guild data. Color Research & Application, 22(1), 11-23.
    doi:10.1002/(SICI)1520-6378(199702)22:1<11::AID-COL4>3.0.CO;2-7
"""

from __future__ import annotations

import numpy as np
from functools import partial

from colour.colorimetry import CCS_ILLUMINANTS
from colour.hints import NDArray
from colour.models.rgb import RGB_Colourspace, gamma_function

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "PRIMARIES_CIE_RGB",
    "WHITEPOINT_NAME_CIE_RGB",
    "CCS_WHITEPOINT_CIE_RGB",
    "MATRIX_CIE_RGB_TO_XYZ",
    "MATRIX_XYZ_TO_CIE_RGB",
    "RGB_COLOURSPACE_CIE_RGB",
]

PRIMARIES_CIE_RGB: NDArray = np.array(
    [
        [0.734742840005998, 0.265257159994002],
        [0.273779033824958, 0.717477700256116],
        [0.166555629580280, 0.008910726182545],
    ]
)
"""
*CIE RGB* colourspace primaries.

Notes
-----
-   *CIE RGB* colourspace primaries were computed using
    :attr:`colour.models.rgb.datasets.cie_rgb.MATRIX_CIE_RGB_TO_XYZ` attribute
    and :func:`colour.primaries_whitepoint` definition.
"""

WHITEPOINT_NAME_CIE_RGB: str = "E"
"""*CIE RGB* colourspace whitepoint name."""

CCS_WHITEPOINT_CIE_RGB: NDArray = CCS_ILLUMINANTS[
    "CIE 1931 2 Degree Standard Observer"
][WHITEPOINT_NAME_CIE_RGB]
"""*CIE RGB* colourspace whitepoint chromaticity coordinates."""

MATRIX_CIE_RGB_TO_XYZ: NDArray = np.array(
    [
        [0.4900, 0.3100, 0.2000],
        [0.1769, 0.8124, 0.0107],
        [0.0000, 0.0099, 0.9901],
    ]
)
"""*CIE RGB* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_CIE_RGB: NDArray = np.linalg.inv(MATRIX_CIE_RGB_TO_XYZ)
"""*CIE XYZ* tristimulus values to *CIE RGB* colourspace matrix."""

RGB_COLOURSPACE_CIE_RGB: RGB_Colourspace = RGB_Colourspace(
    "CIE RGB",
    PRIMARIES_CIE_RGB,
    CCS_WHITEPOINT_CIE_RGB,
    WHITEPOINT_NAME_CIE_RGB,
    MATRIX_CIE_RGB_TO_XYZ,
    MATRIX_XYZ_TO_CIE_RGB,
    partial(gamma_function, exponent=1 / 2.2),
    partial(gamma_function, exponent=2.2),
)
RGB_COLOURSPACE_CIE_RGB.__doc__ = """
*CIE RGB* colourspace.

References
----------
:cite:`Fairman1997`
"""



"""
Viper Log Encoding
==================

Defines the *Viper Log* encoding:

-   :func:`colour.models.log_encoding_ViperLog`
-   :func:`colour.models.log_decoding_ViperLog`

References
----------
-   :cite:`SonyImageworks2012a` : Sony Imageworks. (2012). make.py. Retrieved
    November 27, 2014, from
    https://github.com/imageworks/OpenColorIO-Configs/blob/master/\
nuke-default/make.py
"""

from __future__ import annotations

import numpy as np

from colour.hints import FloatingOrArrayLike, FloatingOrNDArray
from colour.utilities import as_float, from_range_1, to_domain_1

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "log_encoding_ViperLog",
    "log_decoding_ViperLog",
]


def log_encoding_ViperLog(x: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Define the *Viper Log* log encoding curve / opto-electronic transfer
    function.

    Parameters
    ----------
    x
        Linear data :math:`x`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Non-linear data :math:`y`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``x``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``y``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`SonyImageworks2012a`

    Examples
    --------
    >>> log_encoding_ViperLog(0.18)  # doctest: +ELLIPSIS
    0.6360080...
    """

    x = to_domain_1(x)

    y = (1023 + 500 * np.log10(x)) / 1023

    return as_float(from_range_1(y))


def log_decoding_ViperLog(y: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Define the *Viper Log* log decoding curve / electro-optical transfer
    function.

    Parameters
    ----------
    y
        Non-linear data :math:`y`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Linear data :math:`x`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``y``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``x``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`SonyImageworks2012a`

    Examples
    --------
    >>> log_decoding_ViperLog(0.636008067010413)  # doctest: +ELLIPSIS
    0.1799999...
    """

    y = to_domain_1(y)

    x = 10 ** ((1023 * y - 1023) / 500)

    return as_float(from_range_1(x))



"""Showcases reflectance recovery computations using *Otsu et al. (2018)* method."""

import numpy as np

import colour
from colour.utilities import message_box

message_box('"Otsu et al. (2018)" - Reflectance Recovery Computations')

illuminant = colour.SDS_ILLUMINANTS["D65"]

XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
message_box(
    f'Recovering reflectance using "Otsu et al. (2018)" method from given '
    f'"XYZ" tristimulus values:\n\n\tXYZ: {XYZ}'
)
sd = colour.XYZ_to_sd(XYZ, method="Otsu 2018")
print(sd)
print(colour.recovery.XYZ_to_sd_Otsu2018(XYZ))
print(colour.sd_to_XYZ(sd, illuminant=illuminant) / 100)

print("\n")

message_box(
    'Generating a spectral dataset according to the "Otsu et al. (2018) "'
    "method :"
)
XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
reflectances = colour.colorimetry.sds_and_msds_to_msds(
    colour.SDS_COLOURCHECKERS["ColorChecker N Ohta"].values()
).align(colour.recovery.SPECTRAL_SHAPE_OTSU2018)
node_tree = colour.recovery.Tree_Otsu2018(reflectances)
node_tree.optimise()
dataset = node_tree.to_dataset()
print(colour.recovery.XYZ_to_sd_Otsu2018(XYZ, dataset=dataset))



from functools import partial

from colour.hints import (
    Any,
    FloatingOrArrayLike,
    FloatingOrNDArray,
    IntegerOrArrayLike,
    IntegerOrNDArray,
    Literal,
    Union,
)
from colour.utilities import (
    CaseInsensitiveMapping,
    filter_kwargs,
    usage_warning,
    validate_method,
)

from .common import CV_range, legal_to_full, full_to_legal
from .gamma import gamma_function
from .aces import (
    log_encoding_ACESproxy,
    log_decoding_ACESproxy,
    log_encoding_ACEScc,
    log_decoding_ACEScc,
    log_encoding_ACEScct,
    log_decoding_ACEScct,
)
from .arib_std_b67 import oetf_ARIBSTDB67, oetf_inverse_ARIBSTDB67
from .arri_alexa_log_c import log_encoding_ALEXALogC, log_decoding_ALEXALogC
from .blackmagic_design import (
    oetf_BlackmagicFilmGeneration5,
    oetf_inverse_BlackmagicFilmGeneration5,
)
from .canon_log import (
    log_encoding_CanonLog,
    log_decoding_CanonLog,
    log_encoding_CanonLog2,
    log_decoding_CanonLog2,
    log_encoding_CanonLog3,
    log_decoding_CanonLog3,
)
from .cineon import log_encoding_Cineon, log_decoding_Cineon
from .davinci_intermediate import (
    oetf_DaVinciIntermediate,
    oetf_inverse_DaVinciIntermediate,
)
from .dcdm import eotf_inverse_DCDM, eotf_DCDM
from .dicom_gsdf import eotf_inverse_DICOMGSDF, eotf_DICOMGSDF
from .dji_dlog import log_encoding_DJIDLog, log_decoding_DJIDLog
from .exponent import exponent_function_basic, exponent_function_monitor_curve
from .filmic_pro import log_encoding_FilmicPro6, log_decoding_FilmicPro6
from .filmlight_tlog import (
    log_encoding_FilmLightTLog,
    log_decoding_FilmLightTLog,
)
from .gopro import log_encoding_Protune, log_decoding_Protune
from .itur_bt_601 import oetf_BT601, oetf_inverse_BT601
from .itur_bt_709 import oetf_BT709, oetf_inverse_BT709
from .itur_bt_1886 import eotf_inverse_BT1886, eotf_BT1886
from .itur_bt_2020 import eotf_inverse_BT2020, eotf_BT2020
from .st_2084 import eotf_inverse_ST2084, eotf_ST2084
from .itur_bt_2100 import (
    oetf_PQ_BT2100,
    oetf_inverse_PQ_BT2100,
    eotf_PQ_BT2100,
    eotf_inverse_PQ_BT2100,
    ootf_PQ_BT2100,
    ootf_inverse_PQ_BT2100,
    oetf_HLG_BT2100,
    oetf_inverse_HLG_BT2100,
    BT2100_HLG_EOTF_METHODS,
    eotf_HLG_BT2100,
    BT2100_HLG_EOTF_INVERSE_METHODS,
    eotf_inverse_HLG_BT2100,
    BT2100_HLG_OOTF_METHODS,
    ootf_HLG_BT2100,
    BT2100_HLG_OOTF_INVERSE_METHODS,
    ootf_inverse_HLG_BT2100,
)
from .linear import linear_function
from .log import (
    logarithmic_function_basic,
    logarithmic_function_quasilog,
    logarithmic_function_camera,
    log_encoding_Log2,
    log_decoding_Log2,
)
from .panalog import log_encoding_Panalog, log_decoding_Panalog
from .panasonic_vlog import log_encoding_VLog, log_decoding_VLog
from .fujifilm_flog import log_encoding_FLog, log_decoding_FLog
from .nikon_nlog import log_encoding_NLog, log_decoding_NLog
from .pivoted_log import log_encoding_PivotedLog, log_decoding_PivotedLog
from .red_log import (
    log_encoding_REDLog,
    log_decoding_REDLog,
    log_encoding_REDLogFilm,
    log_decoding_REDLogFilm,
    LOG3G10_ENCODING_METHODS,
    LOG3G10_DECODING_METHODS,
    log_encoding_Log3G10,
    log_decoding_Log3G10,
    log_encoding_Log3G12,
    log_decoding_Log3G12,
)
from .rimm_romm_rgb import (
    cctf_encoding_ROMMRGB,
    cctf_decoding_ROMMRGB,
    cctf_encoding_ProPhotoRGB,
    cctf_decoding_ProPhotoRGB,
    cctf_encoding_RIMMRGB,
    cctf_decoding_RIMMRGB,
    log_encoding_ERIMMRGB,
    log_decoding_ERIMMRGB,
)
from .smpte_240m import oetf_SMPTE240M, eotf_SMPTE240M
from .sony_slog import (
    log_encoding_SLog,
    log_decoding_SLog,
    log_encoding_SLog2,
    log_decoding_SLog2,
    log_encoding_SLog3,
    log_decoding_SLog3,
)
from .srgb import eotf_inverse_sRGB, eotf_sRGB
from .viper_log import log_encoding_ViperLog, log_decoding_ViperLog

__all__ = [
    "CV_range",
    "legal_to_full",
    "full_to_legal",
]
__all__ += [
    "gamma_function",
]
__all__ += [
    "log_encoding_ACESproxy",
    "log_decoding_ACESproxy",
    "log_encoding_ACEScc",
    "log_decoding_ACEScc",
    "log_encoding_ACEScct",
    "log_decoding_ACEScct",
]
__all__ += [
    "oetf_ARIBSTDB67",
    "oetf_inverse_ARIBSTDB67",
]
__all__ += [
    "log_encoding_ALEXALogC",
    "log_decoding_ALEXALogC",
]
__all__ += [
    "oetf_BlackmagicFilmGeneration5",
    "oetf_inverse_BlackmagicFilmGeneration5",
]
__all__ += [
    "log_encoding_CanonLog",
    "log_decoding_CanonLog",
    "log_encoding_CanonLog2",
    "log_decoding_CanonLog2",
    "log_encoding_CanonLog3",
    "log_decoding_CanonLog3",
]
__all__ += [
    "log_encoding_Cineon",
    "log_decoding_Cineon",
]
__all__ += [
    "oetf_DaVinciIntermediate",
    "oetf_inverse_DaVinciIntermediate",
]
__all__ += [
    "eotf_inverse_DCDM",
    "eotf_DCDM",
]
__all__ += [
    "eotf_inverse_DICOMGSDF",
    "eotf_DICOMGSDF",
]
__all__ += [
    "log_encoding_DJIDLog",
    "log_decoding_DJIDLog",
]
__all__ += [
    "exponent_function_basic",
    "exponent_function_monitor_curve",
]
__all__ += [
    "log_encoding_FilmicPro6",
    "log_decoding_FilmicPro6",
]
__all__ += [
    "log_encoding_FilmLightTLog",
    "log_decoding_FilmLightTLog",
]
__all__ += [
    "log_encoding_Protune",
    "log_decoding_Protune",
]
__all__ += [
    "oetf_BT601",
    "oetf_inverse_BT601",
]
__all__ += [
    "oetf_BT709",
    "oetf_inverse_BT709",
]
__all__ += [
    "eotf_inverse_BT1886",
    "eotf_BT1886",
]
__all__ += [
    "eotf_inverse_BT2020",
    "eotf_BT2020",
]
__all__ += [
    "eotf_inverse_ST2084",
    "eotf_ST2084",
]
__all__ += [
    "oetf_PQ_BT2100",
    "oetf_inverse_PQ_BT2100",
    "eotf_PQ_BT2100",
    "eotf_inverse_PQ_BT2100",
    "ootf_PQ_BT2100",
    "ootf_inverse_PQ_BT2100",
    "oetf_HLG_BT2100",
    "oetf_inverse_HLG_BT2100",
    "BT2100_HLG_EOTF_METHODS",
    "eotf_HLG_BT2100",
    "BT2100_HLG_EOTF_INVERSE_METHODS",
    "eotf_inverse_HLG_BT2100",
    "BT2100_HLG_OOTF_METHODS",
    "ootf_HLG_BT2100",
    "BT2100_HLG_OOTF_INVERSE_METHODS",
    "ootf_inverse_HLG_BT2100",
]
__all__ += [
    "linear_function",
]
__all__ += [
    "logarithmic_function_basic",
    "logarithmic_function_quasilog",
    "logarithmic_function_camera",
    "log_encoding_Log2",
    "log_decoding_Log2",
]
__all__ += [
    "log_encoding_Panalog",
    "log_decoding_Panalog",
]
__all__ += [
    "log_encoding_VLog",
    "log_decoding_VLog",
]
__all__ += [
    "log_encoding_FLog",
    "log_decoding_FLog",
]
__all__ += [
    "log_encoding_NLog",
    "log_decoding_NLog",
]
__all__ += [
    "log_encoding_PivotedLog",
    "log_decoding_PivotedLog",
]
__all__ += [
    "log_encoding_REDLog",
    "log_decoding_REDLog",
    "log_encoding_REDLogFilm",
    "log_decoding_REDLogFilm",
    "LOG3G10_ENCODING_METHODS",
    "LOG3G10_DECODING_METHODS",
    "log_encoding_Log3G10",
    "log_decoding_Log3G10",
    "log_encoding_Log3G12",
    "log_decoding_Log3G12",
]
__all__ += [
    "cctf_encoding_ROMMRGB",
    "cctf_decoding_ROMMRGB",
    "cctf_encoding_ProPhotoRGB",
    "cctf_decoding_ProPhotoRGB",
    "cctf_encoding_RIMMRGB",
    "cctf_decoding_RIMMRGB",
    "log_encoding_ERIMMRGB",
    "log_decoding_ERIMMRGB",
]
__all__ += [
    "oetf_SMPTE240M",
    "eotf_SMPTE240M",
]
__all__ += [
    "log_encoding_SLog",
    "log_decoding_SLog",
    "log_encoding_SLog2",
    "log_decoding_SLog2",
    "log_encoding_SLog3",
    "log_decoding_SLog3",
]
__all__ += [
    "eotf_inverse_sRGB",
    "eotf_sRGB",
]
__all__ += [
    "log_encoding_ViperLog",
    "log_decoding_ViperLog",
]

LOG_ENCODINGS: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "ACEScc": log_encoding_ACEScc,
        "ACEScct": log_encoding_ACEScct,
        "ACESproxy": log_encoding_ACESproxy,
        "ALEXA Log C": log_encoding_ALEXALogC,
        "Canon Log 2": log_encoding_CanonLog2,
        "Canon Log 3": log_encoding_CanonLog3,
        "Canon Log": log_encoding_CanonLog,
        "Cineon": log_encoding_Cineon,
        "D-Log": log_encoding_DJIDLog,
        "ERIMM RGB": log_encoding_ERIMMRGB,
        "F-Log": log_encoding_FLog,
        "Filmic Pro 6": log_encoding_FilmicPro6,
        "Log2": log_encoding_Log2,
        "Log3G10": log_encoding_Log3G10,
        "Log3G12": log_encoding_Log3G12,
        "N-Log": log_encoding_NLog,
        "PLog": log_encoding_PivotedLog,
        "Panalog": log_encoding_Panalog,
        "Protune": log_encoding_Protune,
        "REDLog": log_encoding_REDLog,
        "REDLogFilm": log_encoding_REDLogFilm,
        "S-Log": log_encoding_SLog,
        "S-Log2": log_encoding_SLog2,
        "S-Log3": log_encoding_SLog3,
        "T-Log": log_encoding_FilmLightTLog,
        "V-Log": log_encoding_VLog,
        "ViperLog": log_encoding_ViperLog,
    }
)
LOG_ENCODINGS.__doc__ = """
Supported *log* encoding functions.
"""


def log_encoding(
    value: FloatingOrArrayLike,
    function: Union[
        Literal[
            "ACEScc",
            "ACEScct",
            "ACESproxy",
            "ALEXA Log C",
            "Canon Log 2",
            "Canon Log 3",
            "Canon Log",
            "Cineon",
            "D-Log",
            "ERIMM RGB",
            "F-Log",
            "Filmic Pro 6",
            "Log2",
            "Log3G10",
            "Log3G12",
            "N-Log",
            "PLog",
            "Panalog",
            "Protune",
            "REDLog",
            "REDLogFilm",
            "S-Log",
            "S-Log2",
            "S-Log3",
            "T-Log",
            "V-Log",
            "ViperLog",
        ],
        str,
    ] = "Cineon",
    **kwargs: Any
) -> Union[FloatingOrNDArray, IntegerOrNDArray]:
    """
    Encode *scene-referred* exposure values to :math:`R'G'B'` video component
    signal value using given *log* encoding function.

    Parameters
    ----------
    value
        *Scene-referred* exposure values.
    function
        *Log* encoding function.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.models.log_encoding_ACEScc`,
        :func:`colour.models.log_encoding_ACEScct`,
        :func:`colour.models.log_encoding_ACESproxy`,
        :func:`colour.models.log_encoding_ALEXALogC`,
        :func:`colour.models.log_encoding_CanonLog2`,
        :func:`colour.models.log_encoding_CanonLog3`,
        :func:`colour.models.log_encoding_CanonLog`,
        :func:`colour.models.log_encoding_Cineon`,
        :func:`colour.models.log_encoding_DJIDLog`,
        :func:`colour.models.log_encoding_ERIMMRGB`,
        :func:`colour.models.log_encoding_FLog`,
        :func:`colour.models.log_encoding_FilmicPro6`,
        :func:`colour.models.log_encoding_Log2`,
        :func:`colour.models.log_encoding_Log3G10`,
        :func:`colour.models.log_encoding_Log3G12`,
        :func:`colour.models.log_encoding_NLog`,
        :func:`colour.models.log_encoding_PivotedLog`,
        :func:`colour.models.log_encoding_Panalog`,
        :func:`colour.models.log_encoding_Protune`,
        :func:`colour.models.log_encoding_REDLog`,
        :func:`colour.models.log_encoding_REDLogFilm`,
        :func:`colour.models.log_encoding_SLog`,
        :func:`colour.models.log_encoding_SLog2`,
        :func:`colour.models.log_encoding_SLog3`,
        :func:`colour.models.log_encoding_FilmLightTLog`,
        :func:`colour.models.log_encoding_VLog`,
        :func:`colour.models.log_encoding_ViperLog`},
        See the documentation of the previously listed definitions.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.integer` or :class:`numpy.ndarray`
        *Log* values.

    Examples
    --------
    >>> log_encoding(0.18)  # doctest: +ELLIPSIS
    0.4573196...
    >>> log_encoding(0.18, function='ACEScc')  # doctest: +ELLIPSIS
    0.4135884...
    >>> log_encoding(0.18, function='PLog', log_reference=400)
    ... # doctest: +ELLIPSIS
    0.3910068...
    >>> log_encoding(0.18, function='S-Log')  # doctest: +ELLIPSIS
    0.3849708...
    """

    function = validate_method(
        function,
        LOG_ENCODINGS,
        '"{0}" "log" encoding function is invalid, it must be one of {1}!',
    )

    callable_ = LOG_ENCODINGS[function]

    return callable_(value, **filter_kwargs(callable_, **kwargs))


LOG_DECODINGS: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "ACEScc": log_decoding_ACEScc,
        "ACEScct": log_decoding_ACEScct,
        "ACESproxy": log_decoding_ACESproxy,
        "ALEXA Log C": log_decoding_ALEXALogC,
        "Canon Log 2": log_decoding_CanonLog2,
        "Canon Log 3": log_decoding_CanonLog3,
        "Canon Log": log_decoding_CanonLog,
        "Cineon": log_decoding_Cineon,
        "D-Log": log_decoding_DJIDLog,
        "ERIMM RGB": log_decoding_ERIMMRGB,
        "F-Log": log_decoding_FLog,
        "Filmic Pro 6": log_decoding_FilmicPro6,
        "Log2": log_decoding_Log2,
        "Log3G10": log_decoding_Log3G10,
        "Log3G12": log_decoding_Log3G12,
        "N-Log": log_decoding_NLog,
        "PLog": log_decoding_PivotedLog,
        "Panalog": log_decoding_Panalog,
        "Protune": log_decoding_Protune,
        "REDLog": log_decoding_REDLog,
        "REDLogFilm": log_decoding_REDLogFilm,
        "S-Log": log_decoding_SLog,
        "S-Log2": log_decoding_SLog2,
        "S-Log3": log_decoding_SLog3,
        "T-Log": log_decoding_FilmLightTLog,
        "V-Log": log_decoding_VLog,
        "ViperLog": log_decoding_ViperLog,
    }
)
LOG_DECODINGS.__doc__ = """
Supported *log* decoding functions.
"""


def log_decoding(
    value: Union[FloatingOrArrayLike, IntegerOrArrayLike],
    function: Union[
        Literal[
            "ACEScc",
            "ACEScct",
            "ACESproxy",
            "ALEXA Log C",
            "Canon Log 2",
            "Canon Log 3",
            "Canon Log",
            "Cineon",
            "D-Log",
            "ERIMM RGB",
            "F-Log",
            "Filmic Pro 6",
            "Log2",
            "Log3G10",
            "Log3G12",
            "N-Log",
            "PLog",
            "Panalog",
            "Protune",
            "REDLog",
            "REDLogFilm",
            "S-Log",
            "S-Log2",
            "S-Log3",
            "T-Log",
            "V-Log",
            "ViperLog",
        ],
        str,
    ] = "Cineon",
    **kwargs: Any
) -> FloatingOrNDArray:
    """
    Decode :math:`R'G'B'` video component signal value to *scene-referred*
    exposure values using given *log* decoding function.

    Parameters
    ----------
    value
        *Log* values.
    function
        *Log* decoding function.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.models.log_decoding_ACEScc`,
        :func:`colour.models.log_decoding_ACEScct`,
        :func:`colour.models.log_decoding_ACESproxy`,
        :func:`colour.models.log_decoding_ALEXALogC`,
        :func:`colour.models.log_decoding_CanonLog2`,
        :func:`colour.models.log_decoding_CanonLog3`,
        :func:`colour.models.log_decoding_CanonLog`,
        :func:`colour.models.log_decoding_Cineon`,
        :func:`colour.models.log_decoding_DJIDLog`,
        :func:`colour.models.log_decoding_ERIMMRGB`,
        :func:`colour.models.log_decoding_FLog`,
        :func:`colour.models.log_decoding_FilmicPro6`,
        :func:`colour.models.log_decoding_Log2`,
        :func:`colour.models.log_decoding_Log3G10`,
        :func:`colour.models.log_decoding_Log3G12`,
        :func:`colour.models.log_decoding_NLog`,
        :func:`colour.models.log_decoding_PivotedLog`,
        :func:`colour.models.log_decoding_Panalog`,
        :func:`colour.models.log_decoding_Protune`,
        :func:`colour.models.log_decoding_REDLog`,
        :func:`colour.models.log_decoding_REDLogFilm`,
        :func:`colour.models.log_decoding_SLog`,
        :func:`colour.models.log_decoding_SLog2`,
        :func:`colour.models.log_decoding_SLog3`,
        :func:`colour.models.log_decoding_FilmLightTLog`,
        :func:`colour.models.log_decoding_VLog`,
        :func:`colour.models.log_decoding_ViperLog`},
        See the documentation of the previously listed definitions.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *Scene-referred* exposure values.

    Examples
    --------
    >>> log_decoding(0.457319613085418)  # doctest: +ELLIPSIS
    0.1...
    >>> log_decoding(0.413588402492442, function='ACEScc')
    ... # doctest: +ELLIPSIS
    0.1...
    >>> log_decoding(0.391006842619746, function='PLog', log_reference=400)
    ... # doctest: +ELLIPSIS
    0.1...
    >>> log_decoding(0.376512722254600, function='S-Log')
    ... # doctest: +ELLIPSIS
    0.1...
    """

    function = validate_method(
        function,
        LOG_DECODINGS,
        '"{0}" "log" decoding function is invalid, it must be one of {1}!',
    )

    callable_ = LOG_DECODINGS[function]

    return callable_(value, **filter_kwargs(callable_, **kwargs))


__all__ += [
    "LOG_ENCODINGS",
    "LOG_DECODINGS",
]
__all__ += [
    "log_encoding",
    "log_decoding",
]

OETFS: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "ARIB STD-B67": oetf_ARIBSTDB67,
        "Blackmagic Film Generation 5": oetf_BlackmagicFilmGeneration5,
        "DaVinci Intermediate": oetf_DaVinciIntermediate,
        "ITU-R BT.2100 HLG": oetf_HLG_BT2100,
        "ITU-R BT.2100 PQ": oetf_PQ_BT2100,
        "ITU-R BT.601": oetf_BT601,
        "ITU-R BT.709": oetf_BT709,
        "SMPTE 240M": oetf_SMPTE240M,
    }
)
OETFS.__doc__ = """
Supported opto-electrical transfer functions (OETFs / OECFs).
"""


def oetf(
    value: FloatingOrArrayLike,
    function: Union[
        Literal[
            "ARIB STD-B67",
            "Blackmagic Film Generation 5",
            "DaVinci Intermediate",
            "ITU-R BT.2100 HLG",
            "ITU-R BT.2100 PQ",
            "ITU-R BT.601",
            "ITU-R BT.709",
            "SMPTE 240M",
        ],
        str,
    ] = "ITU-R BT.709",
    **kwargs: Any
) -> FloatingOrNDArray:
    """
    Encode estimated tristimulus values in a scene to :math:`R'G'B'` video
    component signal value using given opto-electronic transfer function
    (OETF).

    Parameters
    ----------
    value
        Value.
    function
        Opto-electronic transfer function (OETF).

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.models.oetf_ARIBSTDB67`,
        :func:`colour.models.oetf_BlackmagicFilmGeneration5`,
        :func:`colour.models.oetf_DaVinciIntermediate`,
        :func:`colour.models.oetf_HLG_BT2100`,
        :func:`colour.models.oetf_PQ_BT2100`,
        :func:`colour.models.oetf_BT601`,
        :func:`colour.models.oetf_BT709`,
        :func:`colour.models.oetf_SMPTE240M`},
        See the documentation of the previously listed definitions.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        :math:`R'G'B'` video component signal value.

    Examples
    --------
    >>> oetf(0.18)  # doctest: +ELLIPSIS
    0.4090077...
    >>> oetf(0.18, function='ITU-R BT.601')  # doctest: +ELLIPSIS
    0.4090077...
    """

    function = validate_method(
        function, OETFS, '"{0}" "OETF" is invalid, it must be one of {1}!'
    )

    callable_ = OETFS[function]

    return callable_(value, **filter_kwargs(callable_, **kwargs))


OETF_INVERSES: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "ARIB STD-B67": oetf_inverse_ARIBSTDB67,
        "Blackmagic Film Generation 5": oetf_inverse_BlackmagicFilmGeneration5,
        "DaVinci Intermediate": oetf_inverse_DaVinciIntermediate,
        "ITU-R BT.2100 HLG": oetf_inverse_HLG_BT2100,
        "ITU-R BT.2100 PQ": oetf_inverse_PQ_BT2100,
        "ITU-R BT.601": oetf_inverse_BT601,
        "ITU-R BT.709": oetf_inverse_BT709,
    }
)
OETF_INVERSES.__doc__ = """
Supported inverse opto-electrical transfer functions (OETFs / OECFs).
"""


def oetf_inverse(
    value: FloatingOrArrayLike,
    function: Union[
        Literal[
            "ARIB STD-B67",
            "Blackmagic Film Generation 5",
            "DaVinci Intermediate",
            "ITU-R BT.2100 HLG",
            "ITU-R BT.2100 PQ",
            "ITU-R BT.601",
            "ITU-R BT.709",
        ],
        str,
    ] = "ITU-R BT.709",
    **kwargs: Any
) -> FloatingOrNDArray:
    """
    Decode :math:`R'G'B'` video component signal value to tristimulus values
    at the display using given inverse opto-electronic transfer function
    (OETF).

    Parameters
    ----------
    value
        Value.
    function
        Inverse opto-electronic transfer function (OETF).

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.models.oetf_inverse_ARIBSTDB67`,
        :func:`colour.models.oetf_inverse_BlackmagicFilmGeneration5`,
        :func:`colour.models.oetf_inverse_DaVinciIntermediate`,
        :func:`colour.models.oetf_inverse_HLG_BT2100`,
        :func:`colour.models.oetf_inverse_PQ_BT2100`,
        :func:`colour.models.oetf_inverse_BT601`,
        :func:`colour.models.oetf_inverse_BT709`},
        See the documentation of the previously listed definitions.


    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Tristimulus values at the display.

    Examples
    --------
    >>> oetf_inverse(0.409007728864150)  # doctest: +ELLIPSIS
    0.1...
    >>> oetf_inverse(  # doctest: +ELLIPSIS
    ...     0.409007728864150, function='ITU-R BT.601')
    0.1...
    """

    function = validate_method(
        function,
        OETF_INVERSES,
        '"{0}" inverse "OETF" is invalid, it must be one of {1}!',
    )

    callable_ = OETF_INVERSES[function]

    return callable_(value, **filter_kwargs(callable_, **kwargs))


EOTFS: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "DCDM": eotf_DCDM,
        "DICOM GSDF": eotf_DICOMGSDF,
        "ITU-R BT.1886": eotf_BT1886,
        "ITU-R BT.2020": eotf_BT2020,
        "ITU-R BT.2100 HLG": eotf_HLG_BT2100,
        "ITU-R BT.2100 PQ": eotf_PQ_BT2100,
        "SMPTE 240M": eotf_SMPTE240M,
        "ST 2084": eotf_ST2084,
        "sRGB": eotf_sRGB,
    }
)
EOTFS.__doc__ = """
Supported electro-optical transfer functions (EOTFs / EOCFs).
"""


def eotf(
    value: Union[FloatingOrArrayLike, IntegerOrArrayLike],
    function: Union[
        Literal[
            "DCDM",
            "DICOM GSDF",
            "ITU-R BT.1886",
            "ITU-R BT.2020",
            "ITU-R BT.2100 HLG",
            "ITU-R BT.2100 PQ",
            "SMPTE 240M",
            "ST 2084",
            "sRGB",
        ],
        str,
    ] = "ITU-R BT.1886",
    **kwargs: Any
) -> FloatingOrNDArray:
    """
    Decode :math:`R'G'B'` video component signal value to tristimulus values
    at the display using given electro-optical transfer function (EOTF).

    Parameters
    ----------
    value
        Value.
    function
        Electro-optical transfer function (EOTF).

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.models.eotf_DCDM`,
        :func:`colour.models.eotf_DICOMGSDF`,
        :func:`colour.models.eotf_BT1886`,
        :func:`colour.models.eotf_BT2020`,
        :func:`colour.models.eotf_HLG_BT2100`,
        :func:`colour.models.eotf_PQ_BT2100`,
        :func:`colour.models.eotf_SMPTE240M`,
        :func:`colour.models.eotf_ST2084`,
        :func:`colour.models.eotf_sRGB`},
        See the documentation of the previously listed definitions.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Tristimulus values at the display.

    Examples
    --------
    >>> eotf(0.461356129500442)  # doctest: +ELLIPSIS
    0.1...
    >>> eotf(0.409007728864150, function='ITU-R BT.2020')
    ... # doctest: +ELLIPSIS
    0.1...
    >>> eotf(0.182011532850008, function='ST 2084', L_p=1000)
    ... # doctest: +ELLIPSIS
    0.1...
    """

    function = validate_method(
        function, EOTFS, '"{0}" "EOTF" is invalid, it must be one of {1}!'
    )

    callable_ = EOTFS[function]

    return callable_(value, **filter_kwargs(callable_, **kwargs))


EOTF_INVERSES: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "DCDM": eotf_inverse_DCDM,
        "DICOM GSDF": eotf_inverse_DICOMGSDF,
        "ITU-R BT.1886": eotf_inverse_BT1886,
        "ITU-R BT.2020": eotf_inverse_BT2020,
        "ITU-R BT.2100 HLG": eotf_inverse_HLG_BT2100,
        "ITU-R BT.2100 PQ": eotf_inverse_PQ_BT2100,
        "ST 2084": eotf_inverse_ST2084,
        "sRGB": eotf_inverse_sRGB,
    }
)
EOTF_INVERSES.__doc__ = """
Supported inverse electro-optical transfer functions (EOTFs / EOCFs).
"""


def eotf_inverse(
    value: FloatingOrArrayLike,
    function: Union[
        Literal[
            "DCDM",
            "DICOM GSDF",
            "ITU-R BT.1886",
            "ITU-R BT.2020",
            "ITU-R BT.2100 HLG",
            "ITU-R BT.2100 PQ",
            "ST 2084",
            "sRGB",
        ],
        str,
    ] = "ITU-R BT.1886",
    **kwargs
) -> Union[FloatingOrNDArray, IntegerOrNDArray]:
    """
    Encode estimated tristimulus values in a scene to :math:`R'G'B'` video
    component signal value using given inverse electro-optical transfer
    function (EOTF).

    Parameters
    ----------
    value
        Value.
    function
        Inverse electro-optical transfer function (EOTF).

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.models.eotf_inverse_DCDM`,
        :func:`colour.models.eotf_inverse_DICOMGSDF`,
        :func:`colour.models.eotf_inverse_BT1886`,
        :func:`colour.models.eotf_inverse_BT2020`,
        :func:`colour.models.eotf_inverse_HLG_BT2100`,
        :func:`colour.models.eotf_inverse_PQ_BT2100`,
        :func:`colour.models.eotf_inverse_ST2084`,
        :func:`colour.models.eotf_inverse_sRGB`},
        See the documentation of the previously listed definitions.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.integer` or :class:`numpy.ndarray`
        :math:`R'G'B'` video component signal value.

    Examples
    --------
    >>> eotf_inverse(0.11699185725296059)  # doctest: +ELLIPSIS
    0.4090077...
    >>> eotf_inverse(  # doctest: +ELLIPSIS
    ...     0.11699185725296059, function='ITU-R BT.1886')
    0.4090077...
    """

    function = validate_method(
        function,
        EOTF_INVERSES,
        '"{0}" inverse "EOTF" is invalid, it must be one of {1}!',
    )

    callable_ = EOTF_INVERSES[function]

    return callable_(value, **filter_kwargs(callable_, **kwargs))


__all__ += [
    "OETFS",
    "OETF_INVERSES",
    "EOTFS",
    "EOTF_INVERSES",
]
__all__ += [
    "oetf",
    "oetf_inverse",
    "eotf",
    "eotf_inverse",
]

CCTF_ENCODINGS: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "Gamma 2.2": partial(gamma_function, exponent=1 / 2.2),
        "Gamma 2.4": partial(gamma_function, exponent=1 / 2.4),
        "Gamma 2.6": partial(gamma_function, exponent=1 / 2.6),
        "ProPhoto RGB": cctf_encoding_ProPhotoRGB,
        "RIMM RGB": cctf_encoding_RIMMRGB,
        "ROMM RGB": cctf_encoding_ROMMRGB,
    }
)
CCTF_ENCODINGS.update(LOG_ENCODINGS)
CCTF_ENCODINGS.update(OETFS)
CCTF_ENCODINGS.update(EOTF_INVERSES)
CCTF_ENCODINGS.__doc__ = """
Supported encoding colour component transfer functions (Encoding CCTFs), a
collection of the functions defined by :attr:`colour.LOG_ENCODINGS`,
:attr:`colour.OETFS`, :attr:`colour.EOTF_INVERSES` attributes, the
:func:`colour.models.cctf_encoding_ProPhotoRGB`,
:func:`colour.models.cctf_encoding_RIMMRGB`,
:func:`colour.models.cctf_encoding_ROMMRGB` definitions and 3 gamma encoding
functions (1 / 2.2, 1 / 2.4, 1 / 2.6).

Warnings
--------
For *ITU-R BT.2100*, only the inverse electro-optical transfer functions
(EOTFs / EOCFs) are exposed by this attribute, See the
:attr:`colour.OETFS` attribute for the opto-electronic transfer functions
(OETF).
"""


def cctf_encoding(
    value: FloatingOrArrayLike,
    function: Union[
        Literal[
            "ACEScc",
            "ACEScct",
            "ACESproxy",
            "ALEXA Log C",
            "ARIB STD-B67",
            "Blackmagic Film Generation 5",
            "Canon Log 2",
            "Canon Log 3",
            "Canon Log",
            "Cineon",
            "D-Log",
            "DCDM",
            "DICOM GSDF",
            "DaVinci Intermediate",
            "ERIMM RGB",
            "F-Log",
            "Filmic Pro 6",
            "Gamma 2.2",
            "Gamma 2.4",
            "Gamma 2.6",
            "ITU-R BT.1886",
            "ITU-R BT.2020",
            "ITU-R BT.2100 HLG",
            "ITU-R BT.2100 PQ",
            "ITU-R BT.601",
            "ITU-R BT.709",
            "Log2",
            "Log3G10",
            "Log3G12",
            "N-Log",
            "PLog",
            "Panalog",
            "ProPhoto RGB",
            "Protune",
            "REDLog",
            "REDLogFilm",
            "RIMM RGB",
            "ROMM RGB",
            "S-Log",
            "S-Log2",
            "S-Log3",
            "SMPTE 240M",
            "ST 2084",
            "T-Log",
            "V-Log",
            "ViperLog",
            "sRGB",
        ],
        str,
    ] = "sRGB",
    **kwargs: Any
) -> Union[FloatingOrNDArray, IntegerOrNDArray]:
    """
    Encode linear :math:`RGB` values to non-linear :math:`R'G'B'` values using
    given encoding colour component transfer function (Encoding CCTF).

    Parameters
    ----------
    value
        Linear :math:`RGB` values.
    function
        {:attr:`colour.CCTF_ENCODINGS`},
        Encoding colour component transfer function.

    Other Parameters
    ----------------
    kwargs
        Keywords arguments for the relevant encoding *CCTF* of the
        :attr:`colour.CCTF_ENCODINGS` attribute collection.

    Warnings
    --------
    For *ITU-R BT.2100*, only the inverse electro-optical transfer functions
    (EOTFs / EOCFs) are exposed by this definition, See the
    :func:`colour.oetf` definition for the opto-electronic transfer functions
    (OETF).

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Non-linear :math:`R'G'B'` values.

    Examples
    --------
    >>> cctf_encoding(0.18, function='PLog', log_reference=400)
    ... # doctest: +ELLIPSIS
    0.3910068...
    >>> cctf_encoding(0.18, function='ST 2084', L_p=1000)
    ... # doctest: +ELLIPSIS
    0.1820115...
    >>> cctf_encoding(  # doctest: +ELLIPSIS
    ...     0.11699185725296059, function='ITU-R BT.1886')
    0.4090077...
    """

    function = validate_method(
        function,
        CCTF_ENCODINGS,
        '"{0}" encoding "CCTF" is invalid, it must be one of {1}!',
    )

    if "itu-r bt.2100" in function:
        usage_warning(
            'With the "ITU-R BT.2100" method, only the inverse '
            "electro-optical transfer functions (EOTFs / EOCFs) are exposed "
            'by this definition, See the "colour.oetf" definition '
            "for the opto-electronic transfer functions (OETF)."
        )

    callable_ = CCTF_ENCODINGS[function]

    return callable_(value, **filter_kwargs(callable_, **kwargs))


CCTF_DECODINGS: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "Gamma 2.2": partial(gamma_function, exponent=2.2),
        "Gamma 2.4": partial(gamma_function, exponent=2.4),
        "Gamma 2.6": partial(gamma_function, exponent=2.6),
        "ProPhoto RGB": cctf_decoding_ProPhotoRGB,
        "RIMM RGB": cctf_decoding_RIMMRGB,
        "ROMM RGB": cctf_decoding_ROMMRGB,
    }
)
CCTF_DECODINGS.update(LOG_DECODINGS)
CCTF_DECODINGS.update(OETF_INVERSES)
CCTF_DECODINGS.update(EOTFS)
CCTF_DECODINGS.__doc__ = """
Supported decoding colour component transfer functions (Decoding CCTFs), a
collection of the functions defined by :attr:`colour.LOG_DECODINGS`,
:attr:`colour.EOTFS`, :attr:`colour.OETF_INVERSES` attributes, the
:func:`colour.models.cctf_decoding_ProPhotoRGB`,
:func:`colour.models.cctf_decoding_RIMMRGB`,
:func:`colour.models.cctf_decoding_ROMMRGB` definitions and 3 gamma decoding
functions (2.2, 2.4, 2.6).

Warnings
--------
For *ITU-R BT.2100*, only the electro-optical transfer functions
(EOTFs / EOCFs) are exposed by this attribute, See the
:attr:`colour.OETF_INVERSES` attribute for the inverse opto-electronic
transfer functions (OETF).

Notes
-----
-   The order by which this attribute is defined and updated is critically
    important to ensure that *ITU-R BT.2100* definitions are reciprocal.
"""


def cctf_decoding(
    value: Union[FloatingOrArrayLike, IntegerOrArrayLike],
    function: Union[
        Literal[
            "ACEScc",
            "ACEScct",
            "ACESproxy",
            "ALEXA Log C",
            "ARIB STD-B67",
            "Blackmagic Film Generation 5",
            "Canon Log 2",
            "Canon Log 3",
            "Canon Log",
            "Cineon",
            "D-Log",
            "DCDM",
            "DICOM GSDF",
            "DaVinci Intermediate",
            "ERIMM RGB",
            "F-Log",
            "Filmic Pro 6",
            "Gamma 2.2",
            "Gamma 2.4",
            "Gamma 2.6",
            "ITU-R BT.1886",
            "ITU-R BT.2020",
            "ITU-R BT.2100 HLG",
            "ITU-R BT.2100 PQ",
            "ITU-R BT.601",
            "ITU-R BT.709",
            "Log2",
            "Log3G10",
            "Log3G12",
            "N-Log",
            "PLog",
            "Panalog",
            "ProPhoto RGB",
            "Protune",
            "REDLog",
            "REDLogFilm",
            "RIMM RGB",
            "ROMM RGB",
            "S-Log",
            "S-Log2",
            "S-Log3",
            "SMPTE 240M",
            "ST 2084",
            "T-Log",
            "V-Log",
            "ViperLog",
            "sRGB",
        ],
        str,
    ] = "sRGB",
    **kwargs: Any
) -> FloatingOrNDArray:
    """
    Decode non-linear :math:`R'G'B'` values to linear :math:`RGB` values using
    given decoding colour component transfer function (Decoding CCTF).

    Parameters
    ----------
    value
        Non-linear :math:`R'G'B'` values.
    function
        {:attr:`colour.CCTF_DECODINGS`},
        Decoding colour component transfer function.

    Other Parameters
    ----------------
    kwargs
        Keywords arguments for the relevant decoding *CCTF* of the
        :attr:`colour.CCTF_DECODINGS` attribute collection.

    Warnings
    --------
    For *ITU-R BT.2100*, only the electro-optical transfer functions
    (EOTFs / EOCFs) are exposed by this definition, See the
    :func:`colour.oetf_inverse` definition for the inverse opto-electronic
    transfer functions (OETF).

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Linear :math:`RGB` values.

    Examples
    --------
    >>> cctf_decoding(0.391006842619746, function='PLog', log_reference=400)
    ... # doctest: +ELLIPSIS
    0.1...
    >>> cctf_decoding(0.182011532850008, function='ST 2084', L_p=1000)
    ... # doctest: +ELLIPSIS
    0.1...
    >>> cctf_decoding(  # doctest: +ELLIPSIS
    ...     0.461356129500442, function='ITU-R BT.1886')
    0.1...
    """

    function = validate_method(
        function,
        CCTF_DECODINGS,
        '"{0}" decoding "CCTF" is invalid, it must be one of {1}!',
    )

    if "itu-r bt.2100" in function:
        usage_warning(
            'With the "ITU-R BT.2100" method, only the electro-optical '
            "transfer functions (EOTFs / EOCFs) are exposed by this "
            'definition, See the "colour.oetf_inverse" definition '
            "for the inverse opto-electronic transfer functions (OETF)."
        )

    callable_ = CCTF_DECODINGS[function]

    return callable_(value, **filter_kwargs(callable_, **kwargs))


__all__ += [
    "CCTF_ENCODINGS",
    "CCTF_DECODINGS",
]
__all__ += [
    "cctf_encoding",
    "cctf_decoding",
]

OOTFS: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "ITU-R BT.2100 HLG": ootf_HLG_BT2100,
        "ITU-R BT.2100 PQ": ootf_PQ_BT2100,
    }
)
OOTFS.__doc__ = """
Supported opto-optical transfer functions (OOTFs / OOCFs).
"""


def ootf(
    value: FloatingOrArrayLike,
    function: Union[
        Literal["ITU-R BT.2100 HLG", "ITU-R BT.2100 PQ"], str
    ] = "ITU-R BT.2100 PQ",
    **kwargs: Any
) -> FloatingOrNDArray:
    """
    Map relative scene linear light to display linear light using given
    opto-optical transfer function (OOTF / OOCF).

    Parameters
    ----------
    value
        Value.
    function
        Opto-optical transfer function (OOTF / OOCF).

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.models.ootf_HLG_BT2100`,
        :func:`colour.models.ootf_PQ_BT2100`},
        See the documentation of the previously listed definitions.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Luminance of a displayed linear component.

    Examples
    --------
    >>> ootf(0.1)  # doctest: +ELLIPSIS
    779.9883608...
    >>> ootf(0.1, function='ITU-R BT.2100 HLG')  # doctest: +ELLIPSIS
    63.0957344...
    """

    function = validate_method(
        function, OOTFS, '"{0}" "OOTF" is invalid, it must be one of {1}!'
    )

    callable_ = OOTFS[function]

    return callable_(value, **filter_kwargs(callable_, **kwargs))


OOTF_INVERSES: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "ITU-R BT.2100 HLG": ootf_inverse_HLG_BT2100,
        "ITU-R BT.2100 PQ": ootf_inverse_PQ_BT2100,
    }
)
OOTF_INVERSES.__doc__ = """
Supported inverse opto-optical transfer functions (OOTFs / OOCFs).
"""


def ootf_inverse(
    value: FloatingOrArrayLike,
    function: Union[
        Literal["ITU-R BT.2100 HLG", "ITU-R BT.2100 PQ"], str
    ] = "ITU-R BT.2100 PQ",
    **kwargs: Any
) -> FloatingOrNDArray:
    """
    Map relative display linear light to scene linear light using given
    inverse opto-optical transfer function (OOTF / OOCF).

    Parameters
    ----------
    value
        Value.
    function
        Inverse opto-optical transfer function (OOTF / OOCF).

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.models.ootf_inverse_HLG_BT2100`,
        :func:`colour.models.ootf_inverse_PQ_BT2100`},
        See the documentation of the previously listed definitions.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Luminance of scene linear light.

    Examples
    --------
    >>> ootf_inverse(779.988360834115840)  # doctest: +ELLIPSIS
    0.1000000...
    >>> ootf_inverse(  # doctest: +ELLIPSIS
    ...     63.095734448019336, function='ITU-R BT.2100 HLG')
    0.1000000...
    """

    function = validate_method(
        function,
        OOTF_INVERSES,
        '"{0}" inverse "OOTF" is invalid, it must be one of {1}!',
    )

    callable_ = OOTF_INVERSES[function]

    return callable_(value, **filter_kwargs(callable_, **kwargs))


__all__ += [
    "OOTFS",
    "OOTF_INVERSES",
]
__all__ += [
    "ootf",
    "ootf_inverse",
]



"""Defines the unit tests for the :mod:`colour.plotting.tm3018.components` module."""

from __future__ import annotations

import unittest
from matplotlib.pyplot import Axes, Figure

from colour.colorimetry import SDS_ILLUMINANTS
from colour.hints import cast
from colour.quality import (
    ColourQuality_Specification_ANSIIESTM3018,
    colour_fidelity_index_ANSIIESTM3018,
)
from colour.plotting.tm3018.components import (
    plot_spectra_ANSIIESTM3018,
    plot_colour_vector_graphic,
    plot_16_bin_bars,
    plot_local_chroma_shifts,
    plot_local_hue_shifts,
    plot_local_colour_fidelities,
    plot_colour_fidelity_indexes,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestPlotSpectraANSIIESTM3018",
    "TestPlotColourVectorGraphic",
    "TestPlot16BinBars",
    "TestPlotLocalChromaShifts",
    "TestPlotLocalHueShifts",
    "TestPlotLocalColourFidelities",
    "TestPlotColourFidelityIndexes",
]

SPECIFICATION_ANSIIESTM3018: ColourQuality_Specification_ANSIIESTM3018 = cast(
    ColourQuality_Specification_ANSIIESTM3018,
    colour_fidelity_index_ANSIIESTM3018(SDS_ILLUMINANTS["FL2"], True),
)


class TestPlotSpectraANSIIESTM3018(unittest.TestCase):
    """
        Define :func:`colour.plotting.tm3018.components.
    plot_spectra_ANSIIESTM3018` definition unit tests methods.
    """

    def test_plot_spectra_ANSIIESTM3018(self):
        """
        Test :func:`colour.plotting.tm3018.components.\
plot_spectra_ANSIIESTM3018` definition.
        """

        figure, axes = plot_spectra_ANSIIESTM3018(SPECIFICATION_ANSIIESTM3018)

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotColourVectorGraphic(unittest.TestCase):
    """
    Define :func:`colour.plotting.tm3018.components.\
plot_colour_vector_graphic` definition unit tests methods.
    """

    def test_plot_colour_vector_graphic(self):
        """
        Test :func:`colour.plotting.tm3018.components.\
plot_colour_vector_graphic` definition.
        """

        figure, axes = plot_colour_vector_graphic(SPECIFICATION_ANSIIESTM3018)

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlot16BinBars(unittest.TestCase):
    """
    Define :func:`colour.plotting.tm3018.components.plot_16_bin_bars`
    definition unit tests methods.
    """

    def test_plot_16_bin_bars(self):
        """
        Test :func:`colour.plotting.tm3018.components.plot_16_bin_bars`
        definition.
        """

        figure, axes = plot_16_bin_bars(range(16), "{0}")

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotLocalChromaShifts(unittest.TestCase):
    """
    Define :func:`colour.plotting.tm3018.components.plot_local_chroma_shifts`
    definition unit tests methods.
    """

    def test_plot_local_chroma_shifts(self):
        """
        Test :func:`colour.plotting.tm3018.components.\
plot_local_chroma_shifts` definition.
        """

        figure, axes = plot_local_chroma_shifts(SPECIFICATION_ANSIIESTM3018)

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotLocalHueShifts(unittest.TestCase):
    """
    Define :func:`colour.plotting.tm3018.components.plot_local_hue_shifts`
    definition unit tests methods.
    """

    def test_plot_local_hue_shifts(self):
        """
        Test :func:`colour.plotting.tm3018.components.\
plot_local_hue_shifts` definition.
        """

        figure, axes = plot_local_hue_shifts(SPECIFICATION_ANSIIESTM3018)

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotLocalColourFidelities(unittest.TestCase):
    """
        Define :func:`colour.plotting.tm3018.components.
    plot_local_colour_fidelities` definition unit tests methods.
    """

    def test_plot_local_colour_fidelities(self):
        """
        Test :func:`colour.plotting.tm3018.components.\
plot_local_colour_fidelities` definition.
        """

        figure, axes = plot_local_colour_fidelities(
            SPECIFICATION_ANSIIESTM3018
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotColourFidelityIndexes(unittest.TestCase):
    """
    Define :func:`colour.plotting.tm3018.components.\
plot_colour_fidelity_indexes` definition unit tests methods.
    """

    def test_plot_colour_fidelity_indexes(self):
        """
        Test :func:`colour.plotting.tm3018.components.\
plot_colour_fidelity_indexes` definition.
        """

        figure, axes = plot_colour_fidelity_indexes(
            SPECIFICATION_ANSIIESTM3018
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


if __name__ == "__main__":
    unittest.main()



"""
Defines the unit tests for the
:mod:`colour.models.rgb.transfer_functions.common` module.
"""

import numpy as np
import unittest

from colour.models.rgb.transfer_functions import (
    CCTF_DECODINGS,
    CCTF_ENCODINGS,
    EOTFS,
    EOTF_INVERSES,
    LOG_DECODINGS,
    LOG_ENCODINGS,
    OETFS,
    OETF_INVERSES,
    OOTFS,
    OOTF_INVERSES,
    cctf_encoding,
    cctf_decoding,
)
from colour.utilities import ColourUsageWarning, as_int

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Development"

__all__ = [
    "TestCctfEncoding",
    "TestCctfDecoding",
    "TestTransferFunctions",
]


class TestCctfEncoding(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.cctf_encoding`
    definition unit tests methods.
    """

    def test_raise_exception_cctf_encoding(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_encoding_ACESproxy` definition raised exception.
        """

        self.assertWarns(
            ColourUsageWarning,
            cctf_encoding,
            0.18,
            function="ITU-R BT.2100 HLG",
        )
        self.assertWarns(
            ColourUsageWarning,
            cctf_encoding,
            0.18,
            function="ITU-R BT.2100 PQ",
        )


class TestCctfDecoding(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.cctf_decoding`
    definition unit tests methods.
    """

    def test_raise_exception_cctf_decoding(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_encoding_ACESproxy` definition raised exception.
        """

        self.assertWarns(
            ColourUsageWarning,
            cctf_decoding,
            0.18,
            function="ITU-R BT.2100 HLG",
        )
        self.assertWarns(
            ColourUsageWarning,
            cctf_decoding,
            0.18,
            function="ITU-R BT.2100 PQ",
        )


class TestTransferFunctions(unittest.TestCase):
    """Define the transfer functions unit tests methods."""

    def test_transfer_functions(self):
        """Test the transfer functions reciprocity."""

        ignored_transfer_functions = (
            "ACESproxy",
            "DICOM GSDF",
            "Filmic Pro 6",
        )

        decimals = {"D-Log": 1, "F-Log": 4, "N-Log": 3}

        reciprocal_mappings = [
            (LOG_ENCODINGS, LOG_DECODINGS),
            (OETFS, OETF_INVERSES),
            (EOTFS, EOTF_INVERSES),
            (CCTF_ENCODINGS, CCTF_DECODINGS),
            (OOTFS, OOTF_INVERSES),
        ]

        samples = np.hstack(
            [np.linspace(0, 1, as_int(1e5)), np.linspace(0, 65504, 65504 * 10)]
        )

        for encoding_mapping, _decoding_mapping in reciprocal_mappings:
            for name in encoding_mapping:
                if name in ignored_transfer_functions:
                    continue

                encoded_s = CCTF_ENCODINGS[name](samples)
                decoded_s = CCTF_DECODINGS[name](encoded_s)

                np.testing.assert_almost_equal(
                    samples, decoded_s, decimal=decimals.get(name, 7)
                )


if __name__ == "__main__":
    unittest.main()



"""
Zhai and Luo (2018) Chromatic Adaptation Model
==============================================

Defines the *Zhai and Luo (2018)* chromatic adaptation model object:

-   :func:`colour.adaptation.chromatic_adaptation_Zhai2018`

References
----------
-   :cite:`Zhai2018` : Zhai, Q., & Luo, M. R. (2018). Study of chromatic
    adaptation via neutral white matches on different viewing media. Optics
    Express, 26(6), 7724. doi:10.1364/OE.26.007724
"""

import numpy as np

from colour.algebra import vector_dot
from colour.adaptation import CHROMATIC_ADAPTATION_TRANSFORMS
from colour.hints import (
    ArrayLike,
    FloatingOrArrayLike,
    Literal,
    NDArray,
    Union,
)
from colour.utilities import (
    as_float_array,
    from_range_100,
    to_domain_100,
    validate_method,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "chromatic_adaptation_Zhai2018",
]


def chromatic_adaptation_Zhai2018(
    XYZ_b: ArrayLike,
    XYZ_wb: ArrayLike,
    XYZ_wd: ArrayLike,
    D_b: FloatingOrArrayLike = 1,
    D_d: FloatingOrArrayLike = 1,
    XYZ_wo: ArrayLike = np.array([1, 1, 1]),
    transform: Union[Literal["CAT02", "CAT16"], str] = "CAT02",
) -> NDArray:
    """
    Adapt given sample colour :math:`XYZ_{\\beta}` tristimulus values from
    input viewing conditions under :math:`\\beta` illuminant to output viewing
    conditions under :math:`\\delta` illuminant using *Zhai and Luo (2018)*
    chromatic adaptation model.

    According to the definition of :math:`D`, a one-step CAT such as CAT02 can
    only be used to transform colors from an incomplete adapted field into a
    complete adapted field. When CAT02 are used to transform an incomplete to
    incomplete case, :math:`D` has no baseline level to refer to.
    *Smet et al. (2017)* proposed a new concept of two-step CAT to replace the
    present CATs such as CAT02 with only one-step transform in order to define
    :math:`D` more clearly. A two-step CAT involves an illuminant representing
    the baseline states between the test and reference illuminants for the
    calculation. In the first step the test color is transformed from test
    illuminant to the baseline illuminant (:math:`BI`), and it is then
    transformed to the reference illuminant Degrees of adaptation under the
    other illuminants should be calculated relative to the adaptation under the
    :math:`BI`. When :math:`D` becomes lower towards zero, the adaptation point
    of the observer moves towards the :math:`BI`. Therefore, the chromaticity
    of the :math:`BI` should be an intrinsic property of the human vision
    system.

    Parameters
    ----------
    XYZ_b
        Sample colour :math:`XYZ_{\\beta}` under input illuminant
        :math:`\\beta`.
    XYZ_wb
        Input illuminant :math:`\\beta`.
    XYZ_wd
        Output illuminant :math:`\\delta`.
    D_b
        Degree of adaptation :math:`D_{\\beta}` of input illuminant
        :math:`\\beta`.
    D_d
        Degree of adaptation :math:`D_{\\delta}` of output illuminant
        :math:`\\delta`.
    XYZ_wo
        Baseline illuminant (:math:`BI`) :math:`o`.
    transform
        Chromatic adaptation transform.

    Returns
    -------
    :class:`numpy.ndarray`
        Sample corresponding colour :math:`XYZ_{\\delta}` tristimulus values
        under output illuminant :math:`D_{\\delta}`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ_b``  | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+
    | ``XYZ_wb`` | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+
    | ``XYZ_wd`` | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+
    | ``XYZ_wo`` | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ_d``  | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Zhai2018`

    Examples
    --------
    >>> XYZ_b = np.array([48.900, 43.620, 6.250])
    >>> XYZ_wb = np.array([109.850, 100, 35.585])
    >>> XYZ_wd = np.array([95.047, 100, 108.883])
    >>> D_b = 0.9407
    >>> D_d = 0.9800
    >>> XYZ_wo = np.array([100, 100, 100])
    >>> chromatic_adaptation_Zhai2018(
    ...     XYZ_b, XYZ_wb, XYZ_wd, D_b, D_d, XYZ_wo)  # doctest: +ELLIPSIS
    array([ 39.1856164...,  42.1546179...,  19.2367203...])
    >>> XYZ_d = np.array([39.18561644, 42.15461798, 19.23672036])
    >>> chromatic_adaptation_Zhai2018(
    ...     XYZ_d, XYZ_wd, XYZ_wb, D_d, D_b, XYZ_wo)  # doctest: +ELLIPSIS
    array([ 48.9 ,  43.62,   6.25])
    """

    XYZ_b = to_domain_100(XYZ_b)
    XYZ_wb = to_domain_100(XYZ_wb)
    XYZ_wd = to_domain_100(XYZ_wd)
    XYZ_wo = to_domain_100(XYZ_wo)
    D_b = as_float_array(D_b)
    D_d = as_float_array(D_d)

    Y_wb = XYZ_wb[..., 1][..., np.newaxis]
    Y_wd = XYZ_wd[..., 1][..., np.newaxis]
    Y_wo = XYZ_wo[..., 1][..., np.newaxis]

    transform = validate_method(transform, ["CAT02", "CAT16"])
    M = CHROMATIC_ADAPTATION_TRANSFORMS[transform]

    RGB_b = vector_dot(M, XYZ_b)
    RGB_wb = vector_dot(M, XYZ_wb)
    RGB_wd = vector_dot(M, XYZ_wd)
    RGB_wo = vector_dot(M, XYZ_wo)

    D_RGB_b = D_b * (Y_wb / Y_wo) * (RGB_wo / RGB_wb) + 1 - D_b
    D_RGB_d = D_d * (Y_wd / Y_wo) * (RGB_wo / RGB_wd) + 1 - D_d

    D_RGB = D_RGB_b / D_RGB_d

    RGB_d = D_RGB * RGB_b

    XYZ_d = vector_dot(np.linalg.inv(M), RGB_d)

    return from_range_100(XYZ_d)



"""Defines the unit tests for the :mod:`colour.io.luts.common` module."""

import unittest

from colour.io.luts.common import path_to_title

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestPathToTitle",
]


class TestPathToTitle(unittest.TestCase):
    """
    Define :func:`colour.io.luts.common.path_to_title` definition unit tests
    methods.
    """

    def test_path_to_title(self):
        """Test :func:`colour.io.luts.common.path_to_title` definition."""

        self.assertEqual(
            path_to_title(
                "colour/io/luts/tests/resources/cinespace/RGB_1_0.5_0.25.csp"
            ),
            "RGB 1 0 5 0 25",
        )


if __name__ == "__main__":
    unittest.main()



"""Defines the unit tests for the :mod:`colour.colorimetry.dominant` module."""

import numpy as np
import unittest
from itertools import permutations

from colour.colorimetry import (
    MSDS_CMFS,
    CCS_ILLUMINANTS,
    dominant_wavelength,
    complementary_wavelength,
    excitation_purity,
    colorimetric_purity,
)
from colour.colorimetry.dominant import (
    closest_spectral_locus_wavelength,
)
from colour.models import XYZ_to_xy
from colour.utilities import ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestClosestSpectralLocusWavelength",
    "TestDominantWavelength",
    "TestComplementaryWavelength",
    "TestExcitationPurity",
    "TestColorimetricPurity",
]


class TestClosestSpectralLocusWavelength(unittest.TestCase):
    """
    Define :func:`colour.colorimetry.dominant.\
closest_spectral_locus_wavelength` definition unit tests methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        self._xy_s = XYZ_to_xy(
            MSDS_CMFS["CIE 1931 2 Degree Standard Observer"].values
        )

        self._xy_D65 = CCS_ILLUMINANTS["CIE 1931 2 Degree Standard Observer"][
            "D65"
        ]

    def test_closest_spectral_locus_wavelength(self):
        """
        Test :func:`colour.colorimetry.dominant.\
closest_spectral_locus_wavelength` definition.
        """

        xy = np.array([0.54369557, 0.32107944])
        xy_n = self._xy_D65
        i_wl, xy_wl = closest_spectral_locus_wavelength(xy, xy_n, self._xy_s)

        self.assertEqual(i_wl, np.array(256))
        np.testing.assert_almost_equal(
            xy_wl, np.array([0.68354746, 0.31628409]), decimal=7
        )

        xy = np.array([0.37605506, 0.24452225])
        i_wl, xy_wl = closest_spectral_locus_wavelength(xy, xy_n, self._xy_s)

        self.assertEqual(i_wl, np.array(248))
        np.testing.assert_almost_equal(
            xy_wl, np.array([0.45723147, 0.13628148]), decimal=7
        )

    def test_n_dimensional_closest_spectral_locus_wavelength(self):
        """
        Test :func:`colour.colorimetry.dominant.\
closest_spectral_locus_wavelength` definition n-dimensional arrays support.
        """

        xy = np.array([0.54369557, 0.32107944])
        xy_n = self._xy_D65
        i_wl, xy_wl = closest_spectral_locus_wavelength(xy, xy_n, self._xy_s)
        i_wl_r, xy_wl_r = np.array(256), np.array([0.68354746, 0.31628409])
        np.testing.assert_almost_equal(i_wl, i_wl_r)
        np.testing.assert_almost_equal(xy_wl, xy_wl_r, decimal=7)

        xy = np.tile(xy, (6, 1))
        xy_n = np.tile(xy_n, (6, 1))
        i_wl, xy_wl = closest_spectral_locus_wavelength(xy, xy_n, self._xy_s)
        i_wl_r = np.tile(i_wl_r, 6)
        xy_wl_r = np.tile(xy_wl_r, (6, 1))
        np.testing.assert_almost_equal(i_wl, i_wl_r)
        np.testing.assert_almost_equal(xy_wl, xy_wl_r, decimal=7)

        xy = np.reshape(xy, (2, 3, 2))
        xy_n = np.reshape(xy_n, (2, 3, 2))
        i_wl, xy_wl = closest_spectral_locus_wavelength(xy, xy_n, self._xy_s)
        i_wl_r = np.reshape(i_wl_r, (2, 3))
        xy_wl_r = np.reshape(xy_wl_r, (2, 3, 2))
        np.testing.assert_almost_equal(i_wl, i_wl_r)
        np.testing.assert_almost_equal(xy_wl, xy_wl_r, decimal=7)

    @ignore_numpy_errors
    def test_nan_closest_spectral_locus_wavelength(self):
        """
        Test :func:`colour.colorimetry.dominant.\
closest_spectral_locus_wavelength` definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=2))
        for case in cases:
            try:
                closest_spectral_locus_wavelength(case, case, self._xy_s)
            except ValueError:
                pass


class TestDominantWavelength(unittest.TestCase):
    """
    Define :func:`colour.colorimetry.dominant.dominant_wavelength` definition
    unit tests methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        self._xy_D65 = CCS_ILLUMINANTS["CIE 1931 2 Degree Standard Observer"][
            "D65"
        ]

    def test_dominant_wavelength(self):
        """
        Test :func:`colour.colorimetry.dominant.dominant_wavelength`
        definition.
        """

        xy = np.array([0.54369557, 0.32107944])
        xy_n = self._xy_D65
        wl, xy_wl, xy_cwl = dominant_wavelength(xy, xy_n)

        self.assertEqual(wl, np.array(616.0))
        np.testing.assert_almost_equal(
            xy_wl, np.array([0.68354746, 0.31628409]), decimal=7
        )
        np.testing.assert_almost_equal(
            xy_cwl, np.array([0.68354746, 0.31628409]), decimal=7
        )

        xy = np.array([0.37605506, 0.24452225])
        i_wl, xy_wl, xy_cwl = dominant_wavelength(xy, xy_n)

        self.assertEqual(i_wl, np.array(-509.0))
        np.testing.assert_almost_equal(
            xy_wl, np.array([0.45723147, 0.13628148]), decimal=7
        )
        np.testing.assert_almost_equal(
            xy_cwl, np.array([0.01040962, 0.73207453]), decimal=7
        )

    def test_n_dimensional_dominant_wavelength(self):
        """
        Test :func:`colour.colorimetry.dominant.dominant_wavelength`
        definition n-dimensional arrays support.
        """

        xy = np.array([0.54369557, 0.32107944])
        xy_n = self._xy_D65
        wl, xy_wl, xy_cwl = dominant_wavelength(xy, xy_n)
        wl_r, xy_wl_r, xy_cwl_r = (
            np.array(616.0),
            np.array([0.68354746, 0.31628409]),
            np.array([0.68354746, 0.31628409]),
        )
        np.testing.assert_almost_equal(wl, wl_r)
        np.testing.assert_almost_equal(xy_wl, xy_wl_r, decimal=7)
        np.testing.assert_almost_equal(xy_cwl, xy_cwl_r, decimal=7)

        xy = np.tile(xy, (6, 1))
        xy_n = np.tile(xy_n, (6, 1))
        wl, xy_wl, xy_cwl = dominant_wavelength(xy, xy_n)
        wl_r = np.tile(wl_r, 6)
        xy_wl_r = np.tile(xy_wl_r, (6, 1))
        xy_cwl_r = np.tile(xy_cwl_r, (6, 1))
        np.testing.assert_almost_equal(wl, wl_r)
        np.testing.assert_almost_equal(xy_wl, xy_wl_r, decimal=7)
        np.testing.assert_almost_equal(xy_cwl, xy_cwl_r, decimal=7)

        xy = np.reshape(xy, (2, 3, 2))
        xy_n = np.reshape(xy_n, (2, 3, 2))
        wl, xy_wl, xy_cwl = dominant_wavelength(xy, xy_n)
        wl_r = np.reshape(wl_r, (2, 3))
        xy_wl_r = np.reshape(xy_wl_r, (2, 3, 2))
        xy_cwl_r = np.reshape(xy_cwl_r, (2, 3, 2))
        np.testing.assert_almost_equal(wl, wl_r)
        np.testing.assert_almost_equal(xy_wl, xy_wl_r, decimal=7)
        np.testing.assert_almost_equal(xy_cwl, xy_cwl_r, decimal=7)

    @ignore_numpy_errors
    def test_nan_dominant_wavelength(self):
        """
        Test :func:`colour.colorimetry.dominant.dominant_wavelength`
        definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=2))
        for case in cases:
            try:
                dominant_wavelength(case, case)
            except ValueError:
                pass


class TestComplementaryWavelength(unittest.TestCase):
    """
    Define :func:`colour.colorimetry.dominant.complementary_wavelength`
    definition unit tests methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        self._xy_D65 = CCS_ILLUMINANTS["CIE 1931 2 Degree Standard Observer"][
            "D65"
        ]

    def test_complementary_wavelength(self):
        """
        Test :func:`colour.colorimetry.dominant.complementary_wavelength`
        definition.
        """

        xy = np.array([0.54369557, 0.32107944])
        xy_n = self._xy_D65
        wl, xy_wl, xy_cwl = complementary_wavelength(xy, xy_n)

        self.assertEqual(wl, np.array(492.0))
        np.testing.assert_almost_equal(
            xy_wl, np.array([0.03647950, 0.33847127]), decimal=7
        )
        np.testing.assert_almost_equal(
            xy_cwl, np.array([0.03647950, 0.33847127]), decimal=7
        )

        xy = np.array([0.37605506, 0.24452225])
        i_wl, xy_wl, xy_cwl = complementary_wavelength(xy, xy_n)

        self.assertEqual(i_wl, np.array(509.0))
        np.testing.assert_almost_equal(
            xy_wl, np.array([0.01040962, 0.73207453]), decimal=7
        )
        np.testing.assert_almost_equal(
            xy_cwl, np.array([0.01040962, 0.73207453]), decimal=7
        )

    def test_n_dimensional_complementary_wavelength(self):
        """
        Test :func:`colour.colorimetry.dominant.complementary_wavelength`
        definition n-dimensional arrays support.
        """

        xy = np.array([0.54369557, 0.32107944])
        xy_n = self._xy_D65
        wl, xy_wl, xy_cwl = complementary_wavelength(xy, xy_n)
        wl_r, xy_wl_r, xy_cwl_r = (
            np.array(492.0),
            np.array([0.03647950, 0.33847127]),
            np.array([0.03647950, 0.33847127]),
        )
        np.testing.assert_almost_equal(wl, wl_r)
        np.testing.assert_almost_equal(xy_wl, xy_wl_r, decimal=7)
        np.testing.assert_almost_equal(xy_cwl, xy_cwl_r, decimal=7)

        xy = np.tile(xy, (6, 1))
        xy_n = np.tile(xy_n, (6, 1))
        wl, xy_wl, xy_cwl = complementary_wavelength(xy, xy_n)
        wl_r = np.tile(wl_r, 6)
        xy_wl_r = np.tile(xy_wl_r, (6, 1))
        xy_cwl_r = np.tile(xy_cwl_r, (6, 1))
        np.testing.assert_almost_equal(wl, wl_r)
        np.testing.assert_almost_equal(xy_wl, xy_wl_r, decimal=7)
        np.testing.assert_almost_equal(xy_cwl, xy_cwl_r, decimal=7)

        xy = np.reshape(xy, (2, 3, 2))
        xy_n = np.reshape(xy_n, (2, 3, 2))
        wl, xy_wl, xy_cwl = complementary_wavelength(xy, xy_n)
        wl_r = np.reshape(wl_r, (2, 3))
        xy_wl_r = np.reshape(xy_wl_r, (2, 3, 2))
        xy_cwl_r = np.reshape(xy_cwl_r, (2, 3, 2))
        np.testing.assert_almost_equal(wl, wl_r)
        np.testing.assert_almost_equal(xy_wl, xy_wl_r, decimal=7)
        np.testing.assert_almost_equal(xy_cwl, xy_cwl_r, decimal=7)

    @ignore_numpy_errors
    def test_nan_complementary_wavelength(self):
        """
        Test :func:`colour.colorimetry.dominant.complementary_wavelength`
        definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=2))
        for case in cases:
            try:
                complementary_wavelength(case, case)
            except ValueError:
                pass


class TestExcitationPurity(unittest.TestCase):
    """
    Define :func:`colour.colorimetry.dominant.excitation_purity` definition
    unit tests methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        self._xy_D65 = CCS_ILLUMINANTS["CIE 1931 2 Degree Standard Observer"][
            "D65"
        ]

    def test_excitation_purity(self):
        """Test :func:`colour.colorimetry.dominant.excitation_purity` definition."""

        xy = np.array([0.54369557, 0.32107944])
        xy_n = self._xy_D65

        self.assertAlmostEqual(
            excitation_purity(xy, xy_n), 0.622885671878446, places=7
        )

        xy = np.array([0.37605506, 0.24452225])
        self.assertAlmostEqual(
            excitation_purity(xy, xy_n), 0.438347859215887, places=7
        )

    def test_n_dimensional_excitation_purity(self):
        """
        Test :func:`colour.colorimetry.dominant.excitation_purity` definition
        n-dimensional arrays support.
        """

        xy = np.array([0.54369557, 0.32107944])
        xy_n = self._xy_D65
        P_e = excitation_purity(xy, xy_n)

        xy = np.tile(xy, (6, 1))
        xy_n = np.tile(xy_n, (6, 1))
        P_e = np.tile(P_e, 6)
        np.testing.assert_almost_equal(
            excitation_purity(xy, xy_n), P_e, decimal=7
        )

        xy = np.reshape(xy, (2, 3, 2))
        xy_n = np.reshape(xy_n, (2, 3, 2))
        P_e = np.reshape(P_e, (2, 3))
        np.testing.assert_almost_equal(
            excitation_purity(xy, xy_n), P_e, decimal=7
        )

    @ignore_numpy_errors
    def test_nan_excitation_purity(self):
        """
        Test :func:`colour.colorimetry.dominant.excitation_purity` definition
        nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=2))
        for case in cases:
            try:
                excitation_purity(case, case)
            except ValueError:
                pass


class TestColorimetricPurity(unittest.TestCase):
    """
    Define :func:`colour.colorimetry.dominant.colorimetric_purity` definition
    unit tests methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        self._xy_D65 = CCS_ILLUMINANTS["CIE 1931 2 Degree Standard Observer"][
            "D65"
        ]

    def test_colorimetric_purity(self):
        """
        Test :func:`colour.colorimetry.dominant.colorimetric_purity`
        definition.
        """

        xy = np.array([0.54369557, 0.32107944])
        xy_n = self._xy_D65

        self.assertAlmostEqual(
            colorimetric_purity(xy, xy_n), 0.613582813175483, places=7
        )

        xy = np.array([0.37605506, 0.24452225])
        self.assertAlmostEqual(
            colorimetric_purity(xy, xy_n), 0.244307811178847, places=7
        )

    def test_n_dimensional_colorimetric_purity(self):
        """
        Test :func:`colour.colorimetry.dominant.colorimetric_purity`
        definition n-dimensional arrays support.
        """

        xy = np.array([0.54369557, 0.32107944])
        xy_n = self._xy_D65
        P_e = colorimetric_purity(xy, xy_n)

        xy = np.tile(xy, (6, 1))
        xy_n = np.tile(xy_n, (6, 1))
        P_e = np.tile(P_e, 6)
        np.testing.assert_almost_equal(
            colorimetric_purity(xy, xy_n), P_e, decimal=7
        )

        xy = np.reshape(xy, (2, 3, 2))
        xy_n = np.reshape(xy_n, (2, 3, 2))
        P_e = np.reshape(P_e, (2, 3))
        np.testing.assert_almost_equal(
            colorimetric_purity(xy, xy_n), P_e, decimal=7
        )

    @ignore_numpy_errors
    def test_nan_colorimetric_purity(self):
        """
        Test :func:`colour.colorimetry.dominant.colorimetric_purity`
        definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=2))
        for case in cases:
            try:
                colorimetric_purity(case, case)
            except ValueError:
                pass


if __name__ == "__main__":
    unittest.main()
